"""Integration tests for SQLite-backed DataTable loading."""

from pathlib import Path
import sqlite3
import tempfile

from soigia.datatable import DataTable


def test_from_sqlite_db_can_read_real_sqlite_file():
    db_path = None
    handle = tempfile.NamedTemporaryFile(dir=Path.cwd(), suffix=".db", delete=False)
    try:
        db_path = Path(handle.name)
        handle.close()

        connection = sqlite3.connect(str(db_path))
        try:
            connection.execute(
                "CREATE TABLE users (id INTEGER PRIMARY KEY AUTOINCREMENT, name TEXT NOT NULL, age INTEGER NOT NULL)"
            )
            connection.executemany(
                "INSERT INTO users (name, age) VALUES (?, ?)",
                [("alice", 19), ("bob", 24), ("carol", 31)],
            )
            connection.commit()
        finally:
            connection.close()

        table = DataTable.from_sqlite_db(
            "SELECT name, age FROM users ORDER BY id",
            connection_string=f"sqlite:///{db_path.as_posix()}",
        )

        assert isinstance(table, DataTable)
        assert table.to_rows() == [
            {"name": "alice", "age": 19},
            {"name": "bob", "age": 24},
            {"name": "carol", "age": 31},
        ]
    finally:
        if handle and not handle.closed:
            handle.close()
        if db_path is not None and db_path.exists():
            db_path.unlink()



"""Smoke tests for the DataTable unittest-style helpers."""

import unittest

from soigia.datatable import DataTable


class DataTableUnittestSmoke(unittest.TestCase):
    def setUp(self):
        self.table = DataTable({"name": ["alice", "bob"], "age": [19, 24]})

    def test_assert_helpers_run_under_unittest(self):
        self.table.assertEqual(len(self.table), 2)
        self.table.assertIn("name", self.table.columns)
        self.table.assertIsInstance(self.table, DataTable)
        self.table.assertGreater(self.table["age"].max(), 20)

    def test_exception_helpers_run_under_unittest(self):
        with self.table.assertRaises(KeyError):
            _ = self.table["missing_column"]

        with self.table.assertRaisesRegex(ValueError, "bad"):
            raise ValueError("bad value")

        with self.table.assertNoLogs("soigia.unittest.smoke", level="INFO"):
            pass


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""Tests for the custom dataframe helper package."""

from datetime import datetime, date
import logging
import os
import tempfile
import unittest
from contextlib import contextmanager
from pathlib import Path
from unittest.mock import patch

import pandas as pd
import pytest

import soigia
import soigia.datatable as dtpkg
from soigia.datatable import DataTable


def test_df_inherits_from_pandas_dataframe():
    frame = DataTable({"a": [1, 2], "b": [3, 4]})

    assert isinstance(frame, pd.DataFrame)
    assert isinstance(frame, DataTable)


def test_df_constructor_normalizes_string_columns_to_lowercase():
    source = pd.DataFrame({"NAME": ["alice"], "Age": [19]})

    frame = DataTable(source)

    assert list(frame.columns) == ["name", "age"]
    assert frame.iloc[0].to_dict() == {"name": "alice", "age": 19}


def test_df_constructor_normalizes_dict_columns_to_lowercase():
    frame = DataTable({"NAME": ["alice"], "CITY": ["HN"]})

    assert list(frame.columns) == ["name", "city"]


def test_df_allows_attribute_access_and_assignment_for_simple_column_names():
    frame = DataTable(pd.DataFrame({"NAME": ["alice", "bob"], "AGE": [19, 24]}))

    assert frame.name.tolist() == ["alice", "bob"]

    frame.name = 123

    assert frame["name"].tolist() == [123, 123]
    assert frame.name.tolist() == [123, 123]


def test_df_operations_preserve_subclass():
    frame = DataTable({"a": [1, 2], "b": [3, 4]})
    sliced = frame[["a"]]

    assert isinstance(sliced, DataTable)


def test_df_helpers_work():
    frame = DataTable({" A ": ["1", None], "B B": ["2", "3"]})

    cleaned = frame.clean_columns().fill_missing(0).to_numeric()

    assert list(cleaned.columns) == ["a", "b_b"]
    assert cleaned["a"].iloc[1] == 0
    assert cleaned["b_b"].dtype.kind in {"i", "u", "f"}


def test_df_from_records_and_json_work():
    records = [{"name": "alice", "age": 19}, {"name": "bob", "age": 24}]
    with tempfile.NamedTemporaryFile("w", suffix=".json", delete=False, encoding="utf-8") as handle:
        handle.write('[{"name":"alice","age":19},{"name":"bob","age":24}]')
        json_path = handle.name

    try:
        from_records = DataTable.from_records(records)
        from_json = DataTable.from_json(json_path)

        assert from_records.equals(from_json)
        assert list(from_json["name"]) == ["alice", "bob"]
    finally:
        os.unlink(json_path)


def test_df_is_exported_from_soigia():
    assert soigia.datatable.DataTable is DataTable


def test_df_database_helpers_are_exported():
    assert callable(dtpkg.db_connector)
    assert callable(dtpkg.from_sql)
    assert callable(dtpkg.from_mysql_db)
    assert callable(dtpkg.from_maria_db)
    assert callable(dtpkg.from_oracle_db)
    assert callable(dtpkg.from_postgres_db)
    assert callable(dtpkg.from_sqlite_db)


def test_df_has_unittest_style_helpers():
    frame = DataTable()

    instance_methods = [
        "assertEqual",
        "assertNotEqual",
        "assertTrue",
        "assertFalse",
        "assertIs",
        "assertIsNot",
        "assertIsNone",
        "assertIsNotNone",
        "assertIn",
        "assertNotIn",
        "assertIsInstance",
        "assertNotIsInstance",
        "assertIsSubclass",
        "assertNotIsSubclass",
        "assertAlmostEqual",
        "assertNotAlmostEqual",
        "assertGreater",
        "assertGreaterEqual",
        "assertLess",
        "assertLessEqual",
        "assertRaises",
        "assertRaisesRegex",
        "assertWarns",
        "assertWarnsRegex",
        "assertLogs",
        "assertNoLogs",
        "assertMultiLineEqual",
        "assertSequenceEqual",
        "assertListEqual",
        "assertTupleEqual",
        "assertSetEqual",
        "assertDictEqual",
        "assertCountEqual",
        "assertRegex",
        "assertNotRegex",
        "fail",
        "addTypeEqualityFunc",
        "setUp",
        "tearDown",
        "subTest",
        "addCleanup",
        "doCleanups",
        "enterContext",
        "run",
        "debug",
        "defaultTestResult",
        "countTestCases",
        "id",
        "shortDescription",
    ]
    class_methods = ["setUpClass", "tearDownClass", "addClassCleanup", "doClassCleanups", "enterClassContext"]

    for name in instance_methods:
        assert callable(getattr(frame, name))
    for name in class_methods:
        assert callable(getattr(DataTable, name))


def test_df_exposes_chainable_query_methods():
    chain_methods = {"filter", "exclude", "order_by", "get", "count", "exists", "first", "last", "paginate"}

    for name in chain_methods:
        assert callable(getattr(DataTable, name))


def test_base_dataframe_exact_usage():
    df_instance = DataTable()

    assert isinstance(df_instance, DataTable)
    assert df_instance.empty is True
    assert df_instance.objects.count() == df_instance.orm.count() == 0
    assert df_instance.manager.count() == df_instance.objects.count()


def test_df_orm_like_queries_work():
    frame = DataTable(
        {
            "name": ["alice", "bob", "carol", "dave"],
            "age": [19, 24, 31, 24],
            "active": [True, True, False, True],
        }
    )

    qs = frame.objects
    filtered = qs.filter(age__gte=24, name__contains="a").values("name", "age")
    excluded = qs.exclude(active=False).values("name")
    ordered = qs.order_by("-age", "name")
    first = ordered.first()
    last = ordered.last()
    item = qs.get(name="bob")

    assert filtered == [{"name": "carol", "age": 31}, {"name": "dave", "age": 24}]
    assert excluded == [{"name": "alice"}, {"name": "bob"}, {"name": "dave"}]
    assert ordered.values_list("name") == [("carol",), ("bob",), ("dave",), ("alice",)]
    assert isinstance(first, DataTable)
    assert isinstance(last, DataTable)
    assert first["name"].iloc[0] == "carol"
    assert last["name"].iloc[0] == "alice"
    assert item["age"].iloc[0] == 24
    assert qs.count() == 4
    assert qs.exists() is True
    assert qs.none().count() == 0


def test_df_values_helpers_work():
    frame = DataTable({"name": ["alice", "bob", "alice"], "age": [19, 24, 19]})

    values = frame.orm.values()
    names = frame.orm.values_list("name", flat=True)
    pairs = frame.orm.values_list("name", "age")
    distinct_values = values.distinct("name", "age")
    distinct_names = names.distinct()
    distinct_flat_names = frame.objects.values_list("name", flat=True).distinct()

    assert values == [{"name": "alice", "age": 19}, {"name": "bob", "age": 24}, {"name": "alice", "age": 19}]
    assert values.first() == {"name": "alice", "age": 19}
    assert values.last() == {"name": "alice", "age": 19}
    assert values.filter(name="alice") == [{"name": "alice", "age": 19}, {"name": "alice", "age": 19}]
    assert values.exclude(name="alice") == [{"name": "bob", "age": 24}]
    assert values.order_by("name") == [
        {"name": "alice", "age": 19},
        {"name": "alice", "age": 19},
        {"name": "bob", "age": 24},
    ]
    assert distinct_values == [{"name": "alice", "age": 19}, {"name": "bob", "age": 24}]
    assert names == ["alice", "bob", "alice"]
    assert names.first() == "alice"
    assert names.last() == "alice"
    assert distinct_names == ["alice", "bob"]
    assert distinct_flat_names == ["alice", "bob"]
    assert pairs == [("alice", 19), ("bob", 24), ("alice", 19)]


def test_df_distinct_preserves_first_occurrence():
    frame = DataTable({"name": ["a", "b", "a", "c", "b"]})

    distinct_names = frame.objects.values_list("name", flat=True).distinct()

    assert distinct_names == ["a", "b", "c"]


def test_df_group_by_and_aggregates_work():
    frame = DataTable(
        {
            "city": ["hanoi", "hanoi", "saigon", "saigon"],
            "amount": [10, 20, 5, 15],
        }
    )

    grouped = frame.orm.group_by("city").annotate(total_amount=("amount", "sum"), avg_amount=("amount", "mean"))
    stats = frame.orm.sum("amount")

    assert grouped.values() == [
        {"city": "hanoi", "total_amount": 30, "avg_amount": 15.0},
        {"city": "saigon", "total_amount": 20, "avg_amount": 10.0},
    ]
    assert stats == {"amount": 50}


def test_df_paginate_work():
    frame = DataTable({"n": [1, 2, 3, 4, 5]})

    page = frame.orm.paginate(limit=2, offset=2).to_df()

    assert list(page["n"]) == [3, 4]


def test_df_between_and_datetime_lookups_work():
    frame = DataTable(
        {
            "amount": [5, 10, 15, 20],
            "created_at": [
                datetime(2026, 1, 1, 8, 0),
                datetime(2026, 1, 2, 8, 0),
                datetime(2026, 1, 3, 8, 0),
                datetime(2026, 1, 4, 8, 0),
            ],
        }
    )

    between = frame.orm.filter(amount__between=(10, 20))
    by_year = frame.orm.filter(created_at__year=2026)
    by_date = frame.orm.filter(created_at__date=date(2026, 1, 2))

    assert list(between["amount"]) == [10, 15, 20]
    assert list(by_year["amount"]) == [5, 10, 15, 20]
    assert list(by_date["amount"]) == [10]


def test_df_order_by_handles_nulls():
    frame = DataTable({"name": ["b", None, "a"], "score": [2, 1, 3], "rank": [2, 1, 1]})

    ordered = frame.orm.order_by("name").to_df()
    nulls_first = frame.orm.order_by("name", nulls_first=True).to_df()
    nulls_first_explicit = frame.orm.order_by("name", nulls_last=False).to_df()
    multi_nulls = frame.orm.order_by("rank", "name", nulls={"rank": "last", "name": "first"}).to_df()

    assert list(ordered["name"]) == ["a", "b", None]
    assert list(nulls_first["name"]) == [None, "a", "b"]
    assert list(nulls_first_explicit["name"]) == [None, "a", "b"]
    assert list(multi_nulls["name"]) == [None, "a", "b"]


def test_df_random_order_work():
    frame = DataTable({"n": [1, 2, 3, 4, 5]})

    shuffled = frame.orm.order_by("?").to_df()
    random_alias = frame.objects.random_order().to_df()

    assert sorted(shuffled["n"].tolist()) == [1, 2, 3, 4, 5]
    assert sorted(random_alias["n"].tolist()) == [1, 2, 3, 4, 5]


def test_df_aggregate_work():
    frame = DataTable({"amount": [10, 20, 30]})

    stats = frame.orm.aggregate(total=("amount", "sum"), avg=("amount", "mean"), max_value=("amount", "max"))

    assert stats == {"total": 60, "avg": 20.0, "max_value": 30}


def test_df_grouped_aggregate_work():
    frame = DataTable(
        {
            "city": ["hanoi", "hanoi", "saigon", "saigon"],
            "amount": [10, 20, 5, 15],
        }
    )

    grouped = frame.orm.group_by("city").aggregate(total=("amount", "sum"), average=("amount", "mean"))

    assert grouped == {
        "total": [
            {"city": "hanoi", "total": 30},
            {"city": "saigon", "total": 20},
        ],
        "average": [
            {"city": "hanoi", "average": 15.0},
            {"city": "saigon", "average": 10.0},
        ],
    }


def test_df_write_helpers_work():
    frame = DataTable({"name": ["alice", "bob"], "age": [19, 24]})

    created = frame.objects.create(name="carol", age=31)
    appended = frame.objects.append_row(name="dave", age=27)
    bulk_created = frame.objects.bulk_create([{"name": "erin", "age": 22}, {"name": "frank", "age": 29}])
    bulk_updated, bulk_updated_count = frame.objects.bulk_update({"age": 25}, return_count=True, age__gte=20)
    fetched, created_flag = frame.objects.get_or_create(defaults={"age": 31}, name="bob")
    updated, updated_created = frame.objects.update_or_create(defaults={"age": 31}, name="dave")
    updated_count = frame.objects.filter(age__gte=20).update(age=25)
    deleted_count = frame.objects.filter(name="alice").delete()

    assert isinstance(created, DataTable)
    assert created.values.tolist() == [["carol", 31]]
    assert list(appended["name"]) == ["alice", "bob", "dave"]
    assert list(bulk_created["name"]) == ["alice", "bob", "erin", "frank"]
    assert list(bulk_updated["age"]) == [19, 25]
    assert bulk_updated_count == 1
    assert fetched.values.tolist() == [["bob", 24]]
    assert created_flag is False
    assert updated.to_dict(orient="records") == [{"name": "dave", "age": 31}]
    assert updated_created is True
    assert updated_count == 1
    assert deleted_count == 1
    assert frame.objects.save().equals(frame)


def test_df_select_aliases_work():
    frame = DataTable({"name": ["alice", "bob"], "age": [19, 24], "city": ["hn", "sg"]})

    selected = frame.objects.select("name", "age").to_df()
    excluded = frame.objects.exclude_fields("city").to_df()

    assert list(selected.columns) == ["name", "age"]
    assert list(excluded.columns) == ["name", "age"]


def test_df_unittest_style_helpers_work():
    frame = DataTable({"n": [1, 2, 3]})

    frame.assertEqual(1, 1)
    frame.assertNotEqual(1, 2)
    frame.assertTrue(True)
    frame.assertFalse(False)
    frame.assertIs(None, None)
    frame.assertIsNot(None, object())
    frame.assertIsNone(None)
    frame.assertIsNotNone(1)
    frame.assertIn("a", "cat")
    frame.assertNotIn("z", "cat")
    frame.assertIsInstance(frame, DataTable)
    frame.assertNotIsInstance(1, DataTable)
    frame.assertIsSubclass(DataTable, pd.DataFrame)
    frame.assertNotIsSubclass(dict, pd.DataFrame)
    frame.assertAlmostEqual(1.0, 1.0001, places=3)
    frame.assertNotAlmostEqual(1.0, 1.2, places=1)
    frame.assertGreater(3, 2)
    frame.assertGreaterEqual(3, 3)
    frame.assertLess(2, 3)
    frame.assertLessEqual(3, 3)
    frame.assertMultiLineEqual("a\nb", "a\nb")
    frame.assertSequenceEqual([1, 2], [1, 2])
    frame.assertListEqual([1, 2], [1, 2])
    frame.assertTupleEqual((1, 2), (1, 2))
    frame.assertSetEqual({1, 2}, {2, 1})
    frame.assertDictEqual({"a": 1}, {"a": 1})
    frame.assertCountEqual([1, 2, 2], [2, 1, 2])
    frame.assertRegex("abc", r"b")
    frame.assertNotRegex("abc", r"z")
    frame.assertEqual(frame.countTestCases(), 1)
    assert isinstance(frame.defaultTestResult(), unittest.TestResult)
    assert isinstance(frame.run(), unittest.TestResult)
    assert frame.shortDescription() is None
    assert isinstance(frame.id(), str)

    with frame.assertRaises(ZeroDivisionError):
        1 / 0

    with frame.assertRaisesRegex(ValueError, "bad"):
        raise ValueError("bad value")

    with frame.assertNoLogs("soigia.testcase", level="INFO"):
        pass

    logger_name = "soigia.testcase"
    logger = logging.getLogger(logger_name)
    with frame.assertLogs(logger_name, level="INFO") as captured:
        logger.info("hello")
    assert "hello" in "\n".join(captured.output)

    events = []

    @contextmanager
    def marker():
        events.append("enter")
        try:
            yield "value"
        finally:
            events.append("exit")

    assert frame.enterContext(marker()) == "value"
    frame.addCleanup(events.append, "cleanup")
    frame.doCleanups()
    assert events == ["enter", "cleanup", "exit"]

    class_seen = []
    DataTable.addClassCleanup(class_seen.append, "class-cleanup")
    DataTable.doClassCleanups()
    assert class_seen == ["class-cleanup"]


def test_df_file_and_sql_helpers_work():
    frame = DataTable({"name": ["alice"], "age": [19]})

    assert frame.to_rows() == [{"name": "alice", "age": 19}]
    assert "alice" in frame.to_markdown_table()
    assert "<table" in frame.to_html_table()
    assert "name,age" in frame.to_csv_table()

    rich_table = frame.to_rich_table()
    assert len(rich_table.rows) == 1

    with patch("pandas.read_excel", return_value=pd.DataFrame({"name": ["excel"]})) as mocked_excel:
        excel_df = DataTable.from_excel("people.xlsx")
        assert isinstance(excel_df, DataTable)
        assert list(excel_df["name"]) == ["excel"]
        mocked_excel.assert_called_once_with("people.xlsx")

    with patch("pandas.read_parquet", return_value=pd.DataFrame({"name": ["parquet"]})) as mocked_parquet:
        parquet_df = DataTable.from_parquet("people.parquet")
        assert isinstance(parquet_df, DataTable)
        assert list(parquet_df["name"]) == ["parquet"]
        mocked_parquet.assert_called_once_with("people.parquet")

    with patch("pandas.read_feather", return_value=pd.DataFrame({"name": ["feather"]})) as mocked_feather:
        feather_df = DataTable.from_feather("people.feather")
        assert isinstance(feather_df, DataTable)
        assert list(feather_df["name"]) == ["feather"]
        mocked_feather.assert_called_once_with("people.feather")

    with patch("pandas.read_pickle", return_value=pd.DataFrame({"name": ["pickle"]})) as mocked_pickle:
        pickle_df = DataTable.from_pickle("people.pkl")
        assert isinstance(pickle_df, DataTable)
        assert list(pickle_df["name"]) == ["pickle"]
        mocked_pickle.assert_called_once_with("people.pkl")

    with patch("soigia.datatable.from_sql", return_value=pd.DataFrame({"name": ["sql"]})) as mocked_sql:
        sql_df = DataTable.from_sql("SELECT * FROM users", backend="sqlite")
        assert isinstance(sql_df, DataTable)
        assert list(sql_df["name"]) == ["sql"]
        mocked_sql.assert_called_once_with("SELECT * FROM users", backend="sqlite", connection_string=None)

    with patch("pandas.DataFrame.to_excel") as mocked_to_excel:
        out = frame.to_excel_table("table.xlsx", style_header=False, autofit=False, freeze_header=False)
        assert str(out).endswith("table.xlsx")
        mocked_to_excel.assert_called_once()


def test_df_csv_parquet_and_db_save_load_roundtrip(monkeypatch):
    import sqlite3

    sandbox = Path.cwd() / ".test_artifacts" / "datatable_roundtrip"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)

    frame = DataTable({"name": ["alice"], "age": [19]})

    csv_path = frame.save_to_csv("people.csv")
    assert csv_path.exists()
    csv_loaded = DataTable.load_from_csv(csv_path)
    assert csv_loaded.equals(frame)

    parquet_path = frame.save_to_parquet("people.parquet")
    assert parquet_path.exists()
    parquet_loaded = DataTable.load_from_parquet(parquet_path)
    assert parquet_loaded.equals(frame)

    db_path = sandbox / "people.sqlite3"
    frame.save_to_db("people", connection_string=f"sqlite:///{db_path.as_posix()}")

    db_loaded = DataTable.load_from_db(
        "SELECT * FROM people",
        connection_string=f"sqlite:///{db_path.as_posix()}",
    )
    assert db_loaded.equals(frame)

    with sqlite3.connect(db_path) as conn:
        rows = conn.execute("SELECT COUNT(*) FROM people").fetchone()[0]
    assert rows == 1


def test_df_db_helpers_use_default_sqlite_when_connection_is_omitted(monkeypatch):
    sandbox = Path.cwd() / ".test_artifacts" / "datatable_default_db"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)
    monkeypatch.setattr(dtpkg, "_DEFAULT_CONNECTOR", None)

    frame = DataTable({"name": ["alice"], "age": [19]})

    saved = frame.save_to_db("people")
    assert saved is frame

    default_db = sandbox / "data" / "soigia.sqlite3"
    assert default_db.exists()

    loaded = DataTable.load_from_db("select * from people")
    assert loaded.equals(frame)


def test_df_db_helpers_respect_env_default_sqlite_path(monkeypatch):
    sandbox = Path.cwd() / ".test_artifacts" / "datatable_env_default_db"
    sandbox.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(sandbox)
    monkeypatch.setattr(dtpkg, "_DEFAULT_CONNECTOR", None)

    custom_db = sandbox / "custom" / "env.sqlite3"
    monkeypatch.setenv("SOIGIA_DB_PATH", str(custom_db))

    frame = DataTable({"name": ["alice"], "age": [19]})
    frame.save_to_db("people")

    assert custom_db.exists()
    loaded = DataTable.load_from_db("select * from people")
    assert loaded.equals(frame)


def test_df_render_table_work():
    frame = DataTable({"name": ["alice"], "age": [19]})

    assert "alice" in frame.render_table("markdown")
    assert "<table" in frame.render_table("html")
    assert "name,age" in frame.render_table("csv")
    assert frame.objects.render_table("markdown").startswith("|")


def test_df_display_and_logging_helpers_work(capsys, caplog):
    from io import StringIO

    frame = DataTable({"name": ["alice"], "age": [19]})
    buffer = StringIO()

    text = frame.to_display_text()
    printed = frame.print_table()
    with caplog.at_level(logging.INFO, logger="soigia.datatable.display"):
        logged = frame.log_table(logger="soigia.datatable.display")
        wrapped_logged = frame.objects.log_table(logger="soigia.datatable.display")

    captured = capsys.readouterr()

    assert text == printed == frame.objects.to_display_text()
    assert "alice" in text
    assert text in captured.out
    assert logged == text
    assert wrapped_logged == text
    assert text in caplog.text
    assert frame.objects.print_table(file=buffer) == text
    assert buffer.getvalue().startswith("|")


def test_df_latest_and_earliest_work():
    frame = DataTable(
        {
            "name": ["alice", "bob", "carol"],
            "age": [19, 24, 31],
            "created_at": [
                datetime(2026, 1, 1, 8, 0),
                datetime(2026, 1, 2, 8, 0),
                datetime(2026, 1, 3, 8, 0),
            ],
        }
    )

    latest_age = frame.objects.latest("age")
    earliest_age = frame.objects.earliest("age")
    latest_created = frame.objects.latest("created_at")
    earliest_created = frame.objects.earliest("created_at")

    assert latest_age["name"].iloc[0] == "carol"
    assert earliest_age["name"].iloc[0] == "alice"
    assert latest_created["name"].iloc[0] == "carol"
    assert earliest_created["name"].iloc[0] == "alice"


def test_df_versioning_work():
    frame = DataTable({"name": ["alice", "bob"], "age": [19, None], "city": ["hn", "sg"]})

    frame.init_versions("initial")
    updated = frame.fill_missing(0, columns=["age"])
    updated.mark_version("after_fill")
    trimmed = updated.exclude_fields("city")
    trimmed.mark_version("after_trim")

    history = trimmed.version_history(include_data=True)
    assert trimmed.version_count() == 3
    assert history[0]["label"] == "initial"
    assert history[1]["label"] == "after_fill"
    assert history[2]["label"] == "after_trim"
    assert trimmed.current_version()["label"] == "after_trim"
    initial_rows = trimmed.get_version(0).to_rows()
    filled_rows = trimmed.get_version("after_fill").to_rows()
    restored_rows = trimmed.restore_version("after_fill").to_rows()

    assert initial_rows[0]["age"] == 19.0
    assert pd.isna(initial_rows[1]["age"])
    assert filled_rows[1]["age"] == 0.0
    assert restored_rows[1]["age"] == 0.0

    row_diff = trimmed.diff_versions(0, 1)
    cell_diff = trimmed.diff_cells(0, 1)
    column_diff = trimmed.diff_columns(1, 2)
    summary = trimmed.diff_summary(0, 2)

    assert row_diff["row_delta"] == 0
    assert len(row_diff["removed_rows"]) == 1
    assert len(row_diff["added_rows"]) == 1
    assert len(cell_diff["changed_cells"]) == 1
    assert cell_diff["changed_cells"][0]["column"] == "age"
    assert cell_diff["changed_cells"][0]["new"] == 0
    assert column_diff["removed_columns"] == ["city"]
    assert column_diff["added_columns"] == []
    assert summary["removed_columns"] == ["city"]
    assert summary["changed_cells"] == 3
    assert summary["changed_rows"] == 2


def test_df_compare_frames_work():
    left = DataTable({"name": ["alice", "bob"], "age": [19, 24], "city": ["hn", "sg"]})
    right = DataTable({"name": ["alice", "bob", "carol"], "age": [19, 25, 31], "city": ["hn", "sg", "dn"]})

    exact_same = left.same_rows(right)
    exact_same_alias = left.same(right)
    exact_diff = left.different_rows(right)
    keyed = left.compare_frames(right, key_columns=["name"])

    assert exact_same.to_rows() == [{"name": "alice", "age": 19, "city": "hn"}]
    assert exact_same_alias.to_rows() == exact_same.to_rows()
    assert exact_diff["left_only_count"] == 1
    assert exact_diff["right_only_count"] == 2
    assert exact_diff["changed_count"] == 0
    assert keyed["same_count"] == 1
    assert keyed["left_only_count"] == 0
    assert keyed["right_only_count"] == 1
    assert keyed["changed_count"] == 1
    assert keyed["changed_rows"][0]["key"] == ("bob",)
    assert keyed["changed_rows"][0]["changed_columns"] == ["age"]
    assert left.diff_rows(right, key_columns=["name"])["changed_count"] == 1


def test_df_compare_frames_with_duplicate_keys_uses_latest_row_for_each_key():
    left = DataTable({"id": [1, 1], "value": [10, 20]})
    right = DataTable({"id": [1, 1], "value": [10, 30]})

    comparison = left.compare_frames(right, key_columns=["id"])
    same_rows = left.same(right, key_columns=["id"])
    left_only = left.left_only(right, key_columns=["id"])
    right_only = left.right_only(right, key_columns=["id"])
    differences = left.different_rows(right, key_columns=["id"])
    report = left.diff_report(right, key_columns=["id"])
    cell_diff = left.diff_changed_cells(right, key_columns=["id"])

    assert comparison["changed_count"] == 1
    assert same_rows.shape[0] == 0
    assert left_only.shape[0] == 0
    assert right_only.shape[0] == 0
    assert differences["changed_count"] == 1
    assert differences["changed_rows"][0]["key"] == (1,)
    assert report["changed_count"] == 1
    assert report["changed_cell_count"] == 1
    assert report["has_changes"] is True
    assert comparison["changed_rows"][0]["left"]["value"] == 20
    assert comparison["changed_rows"][0]["right"]["value"] == 30
    assert cell_diff.to_dict(orient="records") == [{"key": (1,), "column": "value", "left": 20, "right": 30}]


def test_df_normalize_and_reconcile_support_cross_source_alignment():
    left = DataTable(
        {
            "order_id": [1, 2],
            "customer_name": [" Alice ", "Bob"],
            "amount": ["10.00", "20.0"],
            "updated_at": ["2026-01-01 10:00:00", "2026-01-02 10:00:00"],
        }
    )
    right = DataTable(
        {
            "order_id": [1, 2],
            "cust_name": ["alice", "bob"],
            "total_amount": [10.04, 19.96],
            "updated_at": ["2026-01-01 10:00:30", "2026-01-02 09:59:45"],
        }
    )

    normalized = left.normalize(rename={"customer_name": "cust_name"}, lower_text=True, collapse_whitespace=True)
    assert normalized.columns.tolist() == ["order_id", "cust_name", "amount", "updated_at"]
    assert normalized["cust_name"].tolist() == ["alice", "bob"]

    report = left.reconcile(
        right,
        key_columns=["order_id"],
        column_map={"cust_name": "customer_name", "total_amount": "amount"},
        lower_text=True,
        numeric_tolerance=0.1,
        datetime_tolerance="45s",
    )

    assert report["same_count"] == 2
    assert report["left_only_count"] == 0
    assert report["right_only_count"] == 0
    assert report["changed_count"] == 0
    assert report["changed_cell_count"] == 0
    assert report["has_changes"] is False


def test_df_reconcile_detects_duplicate_keys():
    left = DataTable({"id": [1, 1], "value": [10, 20]})
    right = DataTable({"id": [1], "value": [10]})

    with pytest.raises(ValueError):
        left.reconcile(right, key_columns=["id"])


def test_df_reconcile_html_report_contains_summary_and_sections():
    left = DataTable({"id": [1, 2], "name": ["alice", "bob"], "amount": [10, 20]})
    right = DataTable({"id": [1, 2, 3], "name": ["alice", "bobby", "carol"], "amount": [10, 25, 30]})

    html = left.reconcile_html_report(
        right,
        key_columns=["id"],
        title="Demo Reconcile Report",
        top_n_mismatched_keys=1,
        include_raw_json=True,
    )
    alias_html = left.to_html_report(
        right,
        key_columns=["id"],
        title="Demo Reconcile Report",
        top_n_mismatched_keys=1,
        include_raw_json=True,
    )

    assert "<!doctype html>" in html.lower()
    assert alias_html == html
    assert "Demo Reconcile Report" in html
    assert "Summary" in html
    assert "Column Differences" in html
    assert "Changed Rows" in html
    assert "Top Mismatched Keys" in html
    assert "Showing top 1" in html
    assert "Only on Left" in html
    assert "Only on Right" in html
    assert "Duplicates" in html
    assert "Match status" in html
    assert 'type="application/json"' in html
    assert "Raw JSON" in html

    output = Path("tests") / "_reconcile-report.html"
    try:
        html_written = left.reconcile_html_report(
            right,
            key_columns=["id"],
            title="Demo Reconcile Report",
            top_n_mismatched_keys=1,
            include_raw_json=True,
            output_path=output,
        )
        assert isinstance(html_written, str)
        assert "Demo Reconcile Report" in html_written
        assert output.exists()
        html_text = output.read_text(encoding="utf-8")
        assert "Demo Reconcile Report" in html_text
        assert "mini-pill" in html_text
        assert "Duplicate rows" in html_text
        assert "Showing top 1" in html_text
        assert "reconcile-report-json" in html_text
    finally:
        if output.exists():
            output.unlink()


def test_df_reconcile_html_report_writes_json_sidecar_and_opens_browser(monkeypatch):
    left = DataTable({"id": [1], "name": ["alice"], "amount": [10]})
    right = DataTable({"id": [1], "name": ["alice"], "amount": [10]})

    html_path = Path("tests") / "_reconcile-report-open.html"
    json_path = Path("tests") / "_reconcile-report.json"
    opened = []

    monkeypatch.setattr(dtpkg.webbrowser, "open", lambda url: opened.append(url) or True)

    try:
        html = left.reconcile_html_report(
            right,
            key_columns=["id"],
            title="Open Demo",
            output_path=html_path,
            download_json_path=json_path,
        )
        assert "Open Demo" in html
        assert html_path.exists()
        assert json_path.exists()
        assert "Open Demo" in html_path.read_text(encoding="utf-8")
        assert '"title": "Open Demo"' in json_path.read_text(encoding="utf-8")

        opened_path = left.open_reconcile_html_report(
            html_path,
            right,
            key_columns=["id"],
            title="Open Demo",
        )
        assert opened_path == html_path
        assert opened
        assert opened[-1].startswith("file:")
    finally:
        if html_path.exists():
            html_path.unlink()
        if json_path.exists():
            json_path.unlink()


def test_df_reconcile_html_report_can_open_without_output_path(monkeypatch):
    left = DataTable({"id": [1], "value": [10]})
    right = DataTable({"id": [1], "value": [10]})
    opened = []

    monkeypatch.setattr(dtpkg.webbrowser, "open", lambda url: opened.append(url) or True)

    html = left.reconcile_html_report(
        right,
        key_columns=["id"],
        open_report_in_browser=True,
        title="Temp Open Demo",
    )

    assert "Temp Open Demo" in html
    assert opened
    assert opened[-1].startswith("file:")


def test_df_reconcile_report_path_returns_generated_file(monkeypatch):
    left = DataTable({"id": [1], "value": [10]})
    right = DataTable({"id": [1], "value": [10]})
    opened = []

    monkeypatch.setattr(dtpkg.webbrowser, "open", lambda url: opened.append(url) or True)

    output = left.reconcile_report_path(right, title="Path Demo", open_report_in_browser=True)
    try:
        assert output.exists()
        assert output.suffix == ".html"
        assert opened
        assert opened[-1].startswith("file:")
        assert "Path Demo" in output.read_text(encoding="utf-8")
    finally:
        if output.exists():
            output.unlink()


def test_df_reconcile_report_json_path_returns_generated_file():
    left = DataTable({"id": [1], "value": [10]})
    right = DataTable({"id": [1], "value": [10]})

    output = left.reconcile_report_json_path(right, title="JSON Demo")
    try:
        assert output.exists()
        assert output.suffix == ".json"
        assert '"title": "JSON Demo"' in output.read_text(encoding="utf-8")
    finally:
        if output.exists():
            output.unlink()


def test_df_open_reconcile_html_report_accepts_missing_path(monkeypatch):
    left = DataTable({"id": [1], "value": [10]})
    right = DataTable({"id": [1], "value": [10]})
    opened = []

    monkeypatch.setattr(dtpkg.webbrowser, "open", lambda url: opened.append(url) or True)

    output = left.open_reconcile_html_report(right, title="No Path Open Demo")
    try:
        assert output.exists()
        assert output.suffix == ".html"
        assert opened
        assert "No Path Open Demo" in output.read_text(encoding="utf-8")
    finally:
        if output.exists():
            output.unlink()


def test_df_reconcile_html_report_can_return_payload():
    left = DataTable({"id": [1], "value": [10]})
    right = DataTable({"id": [1], "value": [10]})

    html, payload = left.reconcile_html_report(right, key_columns=["id"], title="Payload Demo", return_payload=True)

    assert "Payload Demo" in html
    assert payload["summary"]["has_changes"] is False
    assert payload["title"] == "Payload Demo"
    assert payload["key_columns"] == ["id"]


def test_df_reconcile_bundle_returns_html_json_and_paths(monkeypatch):
    left = DataTable({"id": [1], "value": [10]})
    right = DataTable({"id": [1], "value": [10]})
    opened = []

    monkeypatch.setattr(dtpkg.webbrowser, "open", lambda url: opened.append(url) or True)

    bundle = left.reconcile_bundle(right, title="Bundle Demo", open_report_in_browser=True)
    try:
        assert bundle["html_path"].exists()
        assert bundle["json_path"].exists()
        assert "Bundle Demo" in bundle["html"]
        assert bundle["payload"]["title"] == "Bundle Demo"
        assert bundle["opened_in_browser"] is True
        assert opened
        assert opened[-1].startswith("file:")
    finally:
        if bundle["html_path"].exists():
            bundle["html_path"].unlink()
        if bundle["json_path"].exists():
            bundle["json_path"].unlink()


def test_df_reconcile_open_returns_bundle(monkeypatch):
    left = DataTable({"id": [1], "value": [10]})
    right = DataTable({"id": [1], "value": [10]})
    opened = []

    monkeypatch.setattr(dtpkg.webbrowser, "open", lambda url: opened.append(url) or True)

    bundle = left.reconcile_open(right, title="Open Bundle Demo")
    try:
        assert bundle["html_path"].exists()
        assert bundle["json_path"].exists()
        assert bundle["payload"]["title"] == "Open Bundle Demo"
        assert opened
        assert opened[-1].startswith("file:")
    finally:
        if bundle["html_path"].exists():
            bundle["html_path"].unlink()
        if bundle["json_path"].exists():
            bundle["json_path"].unlink()


def test_df_compare_aliases_work_through_queryset_wrapper():
    left = DataTable({"name": ["alice", "bob"], "score": [10, 20]})
    right = DataTable({"name": ["alice", "carol"], "score": [10, 30]})

    assert left.objects.same(right, key_columns=["name"]).to_rows() == [{"name": "alice", "score": 10}]
    assert left.objects.left_only(right, key_columns=["name"]).to_rows() == [{"name": "bob", "score": 20}]
    assert left.objects.right_only(right, key_columns=["name"]).to_rows() == [{"name": "carol", "score": 30}]


def test_df_compare_with_status_work():
    left = DataTable({"name": ["alice", "bob"], "score": [10, 20]})
    right = DataTable({"name": ["alice", "carol"], "score": [10, 30]})

    status_table = left.compare_with_status(right)
    wrapped = left.objects.compare_with_status(right)
    custom = left.compare_with_status(right, status_column="side")
    summary = left.compare_status_summary(right)
    summary_wrapped = left.objects.compare_status_summary(right)
    summary_custom = left.compare_status_summary(right, status_column="side")
    rows = left.compare_status_rows(right)
    rows_wrapped = left.objects.compare_status_rows(right)
    rows_custom = left.compare_status_rows(right, status_column="side")
    sample_rows = left.compare_sample_rows(right)
    sample_rows_wrapped = left.objects.compare_sample_rows(right)
    sample_rows_custom = left.compare_sample_rows(right, status_column="side")

    assert status_table.columns.tolist() == ["name", "score", "status"]
    assert status_table["status"].tolist() == ["same", "left_only", "right_only"]
    assert status_table.to_rows() == wrapped.to_rows()
    assert custom.columns.tolist() == ["name", "score", "side"]
    assert custom["side"].tolist() == ["same", "left_only", "right_only"]
    assert summary.columns.tolist() == ["status", "count", "rows"]
    assert summary["status"].tolist() == ["same", "left_only", "right_only"]
    assert summary["count"].tolist() == [1, 1, 1]
    assert summary.iloc[0]["rows"] == [{"name": "alice", "score": 10}]
    assert summary.iloc[1]["rows"] == [{"name": "bob", "score": 20}]
    assert summary.iloc[2]["rows"] == [{"name": "carol", "score": 30}]
    assert summary.to_rows() == summary_wrapped.to_rows()
    assert summary_custom.columns.tolist() == ["side", "count", "rows"]
    assert summary_custom["side"].tolist() == ["same", "left_only", "right_only"]
    assert rows.columns.tolist() == ["name", "score", "status"]
    assert rows["status"].tolist() == ["same", "left_only", "right_only"]
    assert rows.iloc[0].to_dict() == {"name": "alice", "score": 10, "status": "same"}
    assert rows.iloc[1].to_dict() == {"name": "bob", "score": 20, "status": "left_only"}
    assert rows.iloc[2].to_dict() == {"name": "carol", "score": 30, "status": "right_only"}
    assert rows.to_rows() == rows_wrapped.to_rows()
    assert rows_custom.columns.tolist() == ["name", "score", "side"]
    assert rows_custom["side"].tolist() == ["same", "left_only", "right_only"]
    assert sample_rows.to_rows() == rows.to_rows()
    assert sample_rows.to_rows() == sample_rows_wrapped.to_rows()
    assert sample_rows_custom.columns.tolist() == ["name", "score", "side"]
    assert sample_rows_custom["side"].tolist() == ["same", "left_only", "right_only"]


def test_df_database_classmethods_delegate():
    connector = object()

    with patch("soigia.datatable.db_connector", return_value=connector) as mocked_connector:
        returned = DataTable.db_connector("sqlite:///example.db")
        assert returned is connector
        mocked_connector.assert_called_once_with(connection_string="sqlite:///example.db")

    backend_cases = [
        ("from_mysql_db", "soigia.datatable.from_mysql_db"),
        ("from_maria_db", "soigia.datatable.from_maria_db"),
        ("from_oracle_db", "soigia.datatable.from_oracle_db"),
        ("from_postgres_db", "soigia.datatable.from_postgres_db"),
        ("from_sqlite_db", "soigia.datatable.from_sqlite_db"),
    ]
    for method_name, target in backend_cases:
        with patch(target, return_value=pd.DataFrame({"value": [1]})) as mocked_query:
            result = getattr(DataTable, method_name)("select 1")
            assert isinstance(result, DataTable)
            assert list(result["value"]) == [1]
            mocked_query.assert_called_once_with("select 1", connection_string=None)



from __future__ import annotations

import os
import socket
import subprocess
import sys
import time
from pathlib import Path
import sqlite3
from urllib.error import URLError
from urllib.request import urlopen

import pytest
import pandas as pd

_STREAMLIT = pytest.importorskip("streamlit")


def _find_free_port() -> int:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.bind(("127.0.0.1", 0))
        return int(sock.getsockname()[1])


def _wait_for_http_ok(url: str, timeout_s: float = 15.0) -> bool:
    deadline = time.time() + timeout_s
    last_error: Exception | None = None
    while time.time() < deadline:
        try:
            with urlopen(url, timeout=1.5) as response:
                if response.status == 200:
                    return True
        except (OSError, URLError) as exc:
            last_error = exc
            time.sleep(0.25)
    if last_error is not None:
        raise AssertionError(f"streamlit app did not become ready: {last_error}") from last_error
    return False


@pytest.mark.parametrize(
    "example_name",
    [
        "pipeline_yaml_config_demo.py",
        "ui_account_config.py",
        "ui_csv_import.py",
        "ui_jinja_templates.py",
        "ui_markdown_showcase.py",
        "ui_orders.py",
        "ui_reconciliation.py",
        "ui_users.py",
    ],
)
def test_streamlit_examples_start_and_serve_http(example_name: str):
    example_path = Path(__file__).resolve().parents[1] / "examples" / example_name
    port = _find_free_port()
    env = os.environ.copy()
    env["STREAMLIT_BROWSER_GATHER_USAGE_STATS"] = "false"
    artifact_dir = Path.cwd() / ".streamlit_smoke_artifacts"
    artifact_dir.mkdir(exist_ok=True)
    env["SOIGIA_CONFIG_PATH"] = str((artifact_dir / f"{Path(example_name).stem}.config.yaml").resolve())
    stem = Path(example_name).stem
    stdout_path = artifact_dir / f"{stem}-{port}.stdout.log"
    stderr_path = artifact_dir / f"{stem}-{port}.stderr.log"

    stdout_handle = stdout_path.open("w", encoding="utf-8")
    stderr_handle = stderr_path.open("w", encoding="utf-8")
    proc = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "streamlit",
            "run",
            str(example_path),
            "--server.headless",
            "true",
            "--browser.gatherUsageStats",
            "false",
            "--server.address",
            "127.0.0.1",
            "--server.port",
            str(port),
        ],
        stdout=stdout_handle,
        stderr=stderr_handle,
        env=env,
    )
    try:
        _wait_for_http_ok(f"http://127.0.0.1:{port}")
    finally:
        if proc.poll() is None:
            proc.terminate()
            try:
                proc.wait(timeout=10)
            except subprocess.TimeoutExpired:
                proc.kill()
                proc.wait(timeout=10)
        stdout_handle.close()
        stderr_handle.close()
        stdout_text = stdout_path.read_text(encoding="utf-8", errors="replace")
        stderr_text = stderr_path.read_text(encoding="utf-8", errors="replace")
        stdout_path.unlink(missing_ok=True)
        stderr_path.unlink(missing_ok=True)

    assert proc.returncode is not None
    assert "Traceback" not in stdout_text
    assert "Traceback" not in stderr_text


def test_pipeline_final_template_runs():
    example_path = Path(__file__).resolve().parents[1] / "examples" / "pipeline_final_template.py"
    proc = subprocess.run([sys.executable, "-m", "examples.pipeline_final_template"], capture_output=True, text=True, timeout=20)

    assert proc.returncode == 0
    dump_dir = Path.cwd() / "data" / "SalesPipeline"
    assert (dump_dir / "01_normalize.csv").exists()
    assert (dump_dir / "02_enrich.csv").exists()
    assert (dump_dir / "03_summarize.csv").exists()
    csv_path = dump_dir / "SalesPipeline.csv"
    assert csv_path.exists()
    db_path = dump_dir / "SalesPipeline.sqlite3"
    assert db_path.exists()
    with sqlite3.connect(db_path) as conn:
        row_count = conn.execute("select count(*) from final_data").fetchone()[0]
    assert row_count == 2
    parquet_path = dump_dir / "SalesPipeline.parquet"
    assert parquet_path.exists()
    parquet_df = pd.read_parquet(parquet_path)
    assert parquet_df.shape[0] == 2
    summary_path = dump_dir / "pipeline_summary.md"
    summary_text = summary_path.read_text(encoding="utf-8", errors="replace")
    assert "# Pipeline Run Summary" in summary_text
    assert "| # | Step | Status | Rows | Added Columns | Removed Columns |" in summary_text
    assert "| 01 | normalize | success |" in summary_text
    assert "| 02 | enrich | success |" in summary_text
    assert "| 03 | summarize | success |" in summary_text
    assert "## Model" in summary_text
    assert "CSV Path" in summary_text


def test_pipeline_yaml_config_demo_runs():
    example_path = Path(__file__).resolve().parents[1] / "examples" / "pipeline_yaml_config_demo.py"
    proc = subprocess.run([sys.executable, "-m", "examples.pipeline_yaml_config_demo"], capture_output=True, text=True, timeout=20)

    assert proc.returncode == 0
    dump_dir = Path.cwd() / "data" / "SalesYamlPipeline"
    summary_path = dump_dir / "pipeline_summary.md"
    assert summary_path.exists()
    summary_text = summary_path.read_text(encoding="utf-8", errors="replace")
    assert "sales-yaml-demo" in summary_text
    assert "trim_customers" not in summary_text



"""Mock tests for DataTable DB helpers."""

import pytest

import soigia.datatable as dtpkg
from soigia.datatable import DataTable
from tests.datatable_mock_helpers import FakeEngine


def test_db_connector_and_default_connector_roundtrip(monkeypatch):
    connector = dtpkg.db_connector(
        connection_string="sqlite:///base.db",
        mysql="mysql:///db",
        maria="maria:///db",
        oracle="oracle:///db",
        postgres="postgres:///db",
        sqlite="sqlite:///db",
    )

    assert isinstance(connector, dtpkg.DatabaseConnector)
    assert dtpkg._connector() is connector
    assert connector._engine_url("sqlite") == "sqlite:///db"
    assert connector._engine_url("mysql") == "mysql:///db"
    assert connector._engine_url("postgres") == "postgres:///db"

    monkeypatch.setenv("SOIGIA_ORACLE_URL", "oracle://from-env")
    fresh = dtpkg.DatabaseConnector()
    assert fresh._engine_url("oracle") == "oracle://from-env"

    default_sqlite = dtpkg.DatabaseConnector()._engine_url("sqlite")
    assert default_sqlite.startswith("sqlite:///")


def test_db_engine_and_query_branches(monkeypatch):
    fake_engine = FakeEngine()

    monkeypatch.setattr(dtpkg, "create_engine", lambda url: fake_engine)
    monkeypatch.setattr(dtpkg.pd, "read_sql_query", lambda sql, engine, **kwargs: dtpkg.pd.DataFrame({"sql": [sql]}))

    connector = dtpkg.DatabaseConnector(connection_string="sqlite:///memory.db")

    engine = connector._engine("sqlite")
    assert engine is fake_engine

    table = connector.query("select 1", "sqlite")
    assert isinstance(table, DataTable)
    assert table.to_rows() == [{"sql": "select 1"}]
    assert fake_engine.disposed is True


def test_db_query_disposes_engine_on_exception(monkeypatch):
    fake_engine = FakeEngine()

    monkeypatch.setattr(dtpkg, "create_engine", lambda url: fake_engine)

    def boom(sql, engine, **kwargs):
        raise RuntimeError("query failed")

    monkeypatch.setattr(dtpkg.pd, "read_sql_query", boom)
    connector = dtpkg.DatabaseConnector(connection_string="sqlite:///memory.db")

    with pytest.raises(RuntimeError):
        connector.query("select 1", "sqlite")

    assert fake_engine.disposed is True


def test_db_engine_import_error_branch(monkeypatch):
    monkeypatch.setattr(dtpkg, "create_engine", None)
    connector = dtpkg.DatabaseConnector(connection_string="sqlite:///memory.db")

    with pytest.raises(ImportError):
        connector._engine("sqlite")


def test_db_missing_connection_string_error():
    connector = dtpkg.DatabaseConnector()

    assert connector.from_sql("select 1").iloc[0]["1"] == 1

    with pytest.raises(ValueError):
        connector.from_mysql_db("select 1")


def test_db_from_sql_and_backend_wrappers(monkeypatch):
    calls = []

    def fake_query(self, sql, backend, connection_string=None, **kwargs):
        calls.append((backend, sql, connection_string, kwargs))
        return DataTable({"backend": [backend], "sql": [sql]})

    monkeypatch.setattr(dtpkg.DatabaseConnector, "query", fake_query)
    monkeypatch.setattr(dtpkg, "_DEFAULT_CONNECTOR", None)

    connector = dtpkg.db_connector(connection_string="sqlite:///default.db")
    assert connector.from_sql("select 1").iloc[0]["backend"] == "sqlite"
    assert connector.from_mysql_db("select 1").iloc[0]["backend"] == "mysql"
    assert connector.from_maria_db("select 1").iloc[0]["backend"] == "maria"
    assert connector.from_oracle_db("select 1").iloc[0]["backend"] == "oracle"
    assert connector.from_postgres_db("select 1").iloc[0]["backend"] == "postgres"
    assert connector.from_sqlite_db("select 1").iloc[0]["backend"] == "sqlite"

    monkeypatch.setattr(dtpkg, "_DEFAULT_CONNECTOR", connector)
    assert dtpkg.from_sql("select 1").iloc[0]["backend"] == "sqlite"
    assert dtpkg.from_mysql_db("select 1").iloc[0]["backend"] == "mysql"
    assert dtpkg.from_maria_db("select 1").iloc[0]["backend"] == "maria"
    assert dtpkg.from_oracle_db("select 1").iloc[0]["backend"] == "oracle"
    assert dtpkg.from_postgres_db("select 1").iloc[0]["backend"] == "postgres"
    assert dtpkg.from_sqlite_db("select 1").iloc[0]["backend"] == "sqlite"

    assert len(calls) >= 12


def test_db_classmethod_delegation(monkeypatch):
    monkeypatch.setattr(dtpkg, "from_sql", lambda sql, backend="sqlite", connection_string=None, **kwargs: DataTable({"kind": [backend]}))
    monkeypatch.setattr(dtpkg, "from_mysql_db", lambda sql, connection_string=None, **kwargs: DataTable({"kind": ["mysql"]}))
    monkeypatch.setattr(dtpkg, "from_maria_db", lambda sql, connection_string=None, **kwargs: DataTable({"kind": ["maria"]}))
    monkeypatch.setattr(dtpkg, "from_oracle_db", lambda sql, connection_string=None, **kwargs: DataTable({"kind": ["oracle"]}))
    monkeypatch.setattr(dtpkg, "from_postgres_db", lambda sql, connection_string=None, **kwargs: DataTable({"kind": ["postgres"]}))
    monkeypatch.setattr(dtpkg, "from_sqlite_db", lambda sql, connection_string=None, **kwargs: DataTable({"kind": ["sqlite"]}))

    assert DataTable.from_sql("select 1")["kind"].iloc[0] == "sqlite"
    assert DataTable.from_mysql_db("select 1")["kind"].iloc[0] == "mysql"
    assert DataTable.from_maria_db("select 1")["kind"].iloc[0] == "maria"
    assert DataTable.from_oracle_db("select 1")["kind"].iloc[0] == "oracle"
    assert DataTable.from_postgres_db("select 1")["kind"].iloc[0] == "postgres"
    assert DataTable.from_sqlite_db("select 1")["kind"].iloc[0] == "sqlite"



"""Shared YAML-backed configuration helpers for account credentials."""

from __future__ import annotations

import os
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Mapping, MutableMapping, Optional

try:
    import yaml
except Exception as exc:  # pragma: no cover - dependency should be installed
    yaml = None  # type: ignore[assignment]
    _YAML_IMPORT_ERROR = exc
else:
    _YAML_IMPORT_ERROR = None

CONFIG_ENV_VAR = "SOIGIA_CONFIG_PATH"
DEFAULT_CONFIG_FILENAME = "config.yaml"
DEFAULT_ACCOUNT_TYPES = ("google_sheets", "binance", "mt5", "telegram", "shopee")


def _utc_now_iso() -> str:
    return datetime.now(timezone.utc).replace(microsecond=0).isoformat()


def _require_yaml() -> None:
    if yaml is None:  # pragma: no cover - dependency should be installed
        raise RuntimeError("PyYAML is required to load/save config files") from _YAML_IMPORT_ERROR


def resolve_config_path(path: Optional[os.PathLike[str] | str] = None) -> Path:
    if path is not None:
        return Path(path).expanduser().resolve()
    env_path = os.environ.get(CONFIG_ENV_VAR)
    if env_path:
        return Path(env_path).expanduser().resolve()
    return (Path.cwd() / DEFAULT_CONFIG_FILENAME).resolve()


def _default_account_bucket() -> dict[str, dict[str, Any]]:
    return {account_type: {} for account_type in DEFAULT_ACCOUNT_TYPES}


def _default_meta() -> dict[str, Any]:
    return {
        "created_at": _utc_now_iso(),
        "updated_at": _utc_now_iso(),
        "description": "Shared account configuration for API, secrets, and integrations.",
    }


def default_config() -> dict[str, Any]:
    return {
        "version": 1,
        "meta": _default_meta(),
        "accounts": _default_account_bucket(),
    }


def _normalize_mapping(value: Any) -> dict[str, Any]:
    return dict(value) if isinstance(value, Mapping) else {}


def _normalize_account_entry(entry: Any, *, service: str, name: str) -> dict[str, Any]:
    if not isinstance(entry, Mapping):
        entry = {}
    normalized = {
        "service": service,
        "name": name,
        "enabled": bool(entry.get("enabled", True)),
        "label": str(entry.get("label", name)),
        "credentials": _normalize_mapping(entry.get("credentials")),
        "settings": _normalize_mapping(entry.get("settings")),
        "notes": str(entry.get("notes", "")),
        "updated_at": str(entry.get("updated_at", _utc_now_iso())),
    }
    for key, value in entry.items():
        if key not in normalized and key not in {"credentials", "settings"}:
            normalized[key] = value
    return normalized


def normalize_config(data: Any) -> dict[str, Any]:
    config = default_config()
    if not isinstance(data, Mapping):
        return config

    for key in ("version", "meta"):
        if key in data:
            if key == "meta":
                merged_meta = _default_meta()
                merged_meta.update(_normalize_mapping(data[key]))
                config[key] = merged_meta
            else:
                config[key] = data[key]

    accounts = data.get("accounts", {})
    if isinstance(accounts, Mapping):
        merged_accounts: dict[str, dict[str, Any]] = _default_account_bucket()
        for service, service_accounts in accounts.items():
            service_name = str(service)
            bucket: dict[str, Any] = {}
            if isinstance(service_accounts, Mapping):
                for account_name, entry in service_accounts.items():
                    account_name_str = str(account_name)
                    bucket[account_name_str] = _normalize_account_entry(
                        entry,
                        service=service_name,
                        name=account_name_str,
                    )
            merged_accounts[service_name] = bucket
        config["accounts"] = merged_accounts
    return config


def load_config(path: Optional[os.PathLike[str] | str] = None, *, create: bool = False) -> dict[str, Any]:
    _require_yaml()
    config_path = resolve_config_path(path)
    if not config_path.exists():
        config = default_config()
        if create:
            save_config(config, config_path)
        return config
    with config_path.open("r", encoding="utf-8") as handle:
        loaded = yaml.safe_load(handle) or {}
    return normalize_config(loaded)


def save_config(data: Mapping[str, Any], path: Optional[os.PathLike[str] | str] = None) -> Path:
    _require_yaml()
    config_path = resolve_config_path(path)
    config_path.parent.mkdir(parents=True, exist_ok=True)
    normalized = normalize_config(data)
    normalized["meta"]["updated_at"] = _utc_now_iso()
    temp_path = config_path.with_suffix(config_path.suffix + ".tmp")
    with temp_path.open("w", encoding="utf-8") as handle:
        yaml.safe_dump(normalized, handle, sort_keys=False, allow_unicode=True)
    temp_path.replace(config_path)
    return config_path


def get_accounts(config: Mapping[str, Any], service: str) -> dict[str, Any]:
    accounts = config.get("accounts", {})
    if not isinstance(accounts, Mapping):
        return {}
    bucket = accounts.get(service, {})
    return dict(bucket) if isinstance(bucket, Mapping) else {}


def get_account(config: Mapping[str, Any], service: str, name: str) -> dict[str, Any] | None:
    return get_accounts(config, service).get(name)


def upsert_account(
    config: MutableMapping[str, Any],
    service: str,
    name: str,
    account: Mapping[str, Any],
) -> MutableMapping[str, Any]:
    normalized = normalize_config(config)
    accounts = normalized.setdefault("accounts", _default_account_bucket())
    service_bucket = accounts.setdefault(service, {})
    service_bucket[name] = _normalize_account_entry(account, service=service, name=name)
    normalized["meta"]["updated_at"] = _utc_now_iso()
    config.clear()
    config.update(normalized)
    return config


def delete_account(config: MutableMapping[str, Any], service: str, name: str) -> MutableMapping[str, Any]:
    normalized = normalize_config(config)
    accounts = normalized.setdefault("accounts", _default_account_bucket())
    service_bucket = accounts.setdefault(service, {})
    if name in service_bucket:
        del service_bucket[name]
        normalized["meta"]["updated_at"] = _utc_now_iso()
    config.clear()
    config.update(normalized)
    return config


@dataclass(slots=True)
class ConfigStore:
    path: Path | str | None = None

    def resolved_path(self) -> Path:
        return resolve_config_path(self.path)

    def load(self, *, create: bool = False) -> dict[str, Any]:
        return load_config(self.resolved_path(), create=create)

    def save(self, data: Mapping[str, Any]) -> Path:
        return save_config(data, self.resolved_path())

    def get_accounts(self, service: str) -> dict[str, Any]:
        return get_accounts(self.load(), service)

    def get_account(self, service: str, name: str) -> dict[str, Any] | None:
        return get_account(self.load(), service, name)

    def upsert_account(self, service: str, name: str, account: Mapping[str, Any]) -> Path:
        data = self.load()
        upsert_account(data, service, name, account)
        return self.save(data)

    def delete_account(self, service: str, name: str) -> Path:
        data = self.load()
        delete_account(data, service, name)
        return self.save(data)



"""Single-file declarative UI toolkit built on top of Streamlit.

Goal
----
Users should be able to create a web app by writing one file and a few lines
of code, with minimal framework concepts exposed publicly.

Typical usage:

    from soigia.ui import ui

    users = [{"id": 1, "name": "Alice", "active": True}]

    @ui.page("Customers")
    def customers_page(ctx):
        ui.text("Customer Admin")
        search = ui.text_input("Search")
        ui.markdown("Use this panel to manage customers.")
        ui.table(lambda: [row for row in users if search.value.lower() in row["name"].lower()])
        with ui.form("Customer form", submit_label="Save", on_submit=lambda: save_customer()):
            ui.text_input("Name")
            ui.checkbox("Active")
        with ui.dialog("Help", open_label="Need help?"):
            ui.markdown("This is a lightweight management UI.")

    ui.run()

Example app.py:

    from soigia.ui import ui

    orders = [
        {"id": 1, "customer": "Alice", "amount": 120.0, "status": "paid"},
        {"id": 2, "customer": "Bob", "amount": 88.5, "status": "pending"},
    ]

    @ui.page("Orders")
    def orders_page(ctx):
        ui.text("Order Dashboard")
        keyword = ui.text_input("Search")
        ui.metric("Total orders", len(orders))
        ui.table(
            lambda: [row for row in orders if keyword.value.lower() in row["customer"].lower()]
        )
        with ui.form("Update order", submit_label="Save"):
            ui.text_input("Order ID")
            ui.select("Status", ["paid", "pending", "cancelled"])
        ui.button("Refresh")

    ui.run()

Design
------
- Public API is declarative and opinionated.
- Widgets register themselves when created.
- Streamlit stays behind the curtain.
- The file is self-contained on purpose.
"""

from __future__ import annotations

import inspect
import itertools
import re
from math import ceil
from collections.abc import Callable, Mapping, Sequence
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Any, Optional

import pandas as pd

try:
    import streamlit as st
except Exception:  # pragma: no cover - streamlit should exist, but keep import safe.
    st = None  # type: ignore[assignment]

__all__ = [
    "App",
    "ui",
    "run",
    "rerun",
    "stop",
    "state_get",
    "state_set",
    "state_delete",
    "page",
    "text",
    "markdown",
    "code",
    "divider",
    "metric",
    "text_input",
    "textbox",
    "text_area",
    "textarea",
    "checkbox",
    "select",
    "button",
    "table",
    "json",
    "success",
    "info",
    "warning",
    "error",
    "file_uploader",
    "download_button",
    "form",
    "dialog",
    "sidebar",
    "columns",
    "tabs",
    "section",
    "expander",
]

_ORDER = itertools.count()


def _slugify(value: str) -> str:
    value = value.strip().lower()
    value = re.sub(r"[^a-z0-9]+", "_", value)
    return value.strip("_") or "field"


def _coerce_callable(value: Any, context: "RenderContext") -> Any:
    if not callable(value):
        return value
    try:
        sig = inspect.signature(value)
    except (TypeError, ValueError):
        return value()
    params = list(sig.parameters.values())
    if not params:
        return value()
    if any(param.kind == inspect.Parameter.VAR_POSITIONAL for param in params):
        return value(context)
    if len(params) == 1:
        return value(context)
    return value()


def _invoke_callback(callback: Optional[Callable[..., Any]], context: "RenderContext") -> None:
    if callback is None:
        return
    try:
        sig = inspect.signature(callback)
    except (TypeError, ValueError):
        callback()
        return
    params = list(sig.parameters.values())
    if not params:
        callback()
        return
    if any(param.kind == inspect.Parameter.VAR_POSITIONAL for param in params):
        callback(context)
        return
    if len(params) == 1:
        callback(context)
        return
    callback()


def _make_key(kind: str, label: str, explicit_key: Optional[str], order: int) -> str:
    return explicit_key or f"{kind}:{_slugify(label)}:{order}"


def _container_width_value(use_container_width: bool) -> str:
    return "stretch" if use_container_width else "content"


@dataclass(slots=True)
class RenderContext:
    app: "App"
    st: Any

    @property
    def session_state(self) -> Any:
        if st is not None:
            return st.session_state
        return self.st.session_state


@dataclass(slots=True)
class Node:
    label: str = ""
    key: Optional[str] = None
    help: Optional[str] = None
    visible: bool = True
    order: int = field(default_factory=lambda: next(_ORDER), init=False)
    app: Optional["App"] = field(default=None, init=False, repr=False, compare=False)

    def render(self, context: RenderContext) -> Any:
        raise NotImplementedError


@dataclass(slots=True)
class TextNode(Node):
    value: Any = ""
    kind: str = "text"
    truncate_at: Optional[int] = None
    expandable: bool = False

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return None
        content = _coerce_callable(self.value, context)
        preview = content
        if isinstance(content, str) and self.truncate_at and len(content) > self.truncate_at:
            preview = content[: self.truncate_at].rstrip() + "..."
        if self.expandable and preview != content and isinstance(content, str):
            self._render_value(context, preview)
            with context.st.expander(self.label or "Show more"):
                self._render_value(context, content)
            return content
        self._render_value(context, preview)
        return content

    def _render_value(self, context: RenderContext, content: Any) -> None:
        if self.kind == "title":
            context.st.title(content)
        elif self.kind == "header":
            context.st.header(content)
        elif self.kind == "subheader":
            context.st.subheader(content)
        elif self.kind == "caption":
            context.st.caption(content)
        elif self.kind == "markdown":
            context.st.markdown(content)
        elif self.kind == "code":
            context.st.code(content)
        elif self.kind == "json":
            context.st.json(content)
        elif self.kind == "success":
            context.st.success(content)
        elif self.kind == "info":
            context.st.info(content)
        elif self.kind == "warning":
            context.st.warning(content)
        elif self.kind == "error":
            context.st.error(content)
        else:
            context.st.write(content)


@dataclass(slots=True)
class DividerNode(Node):
    def render(self, context: RenderContext) -> Any:
        if self.visible:
            context.st.divider()
        return None


@dataclass(slots=True)
class MetricNode(Node):
    value: Any = ""
    delta: Any = None
    delta_color: str = "normal"

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return None
        value = _coerce_callable(self.value, context)
        delta = _coerce_callable(self.delta, context)
        context.st.metric(self.label or "", value, delta=delta, delta_color=self.delta_color)
        return value


@dataclass(slots=True)
class TableNode(Node):
    data: Any = None
    use_container_width: bool = True
    height: Optional[int] = None
    editable: bool = False
    key_prefix: Optional[str] = None
    paginate: bool = False
    page_size: int = 100
    show_index: bool = False
    empty_message: str = "No rows to display."

    def _to_frame(self, value: Any) -> pd.DataFrame:
        if isinstance(value, pd.DataFrame):
            return value
        if hasattr(value, "to_df") and callable(value.to_df):
            frame = value.to_df()
            if isinstance(frame, pd.DataFrame):
                return frame
        if isinstance(value, dict):
            return pd.DataFrame(value)
        if isinstance(value, list):
            return pd.DataFrame(value)
        return pd.DataFrame([value])

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return None
        value = _coerce_callable(self.data, context)
        frame = self._to_frame(value)
        if frame.empty:
            context.st.info(self.empty_message)
            return frame
        if self.paginate or len(frame) > self.page_size:
            total_rows = len(frame)
            total_pages = max(1, ceil(total_rows / max(1, self.page_size)))
            page_state_key = self.key or self.key_prefix or f"table_page:{self.order}"
            page_widget_key = f"{page_state_key}:widget"
            current_page = int(context.session_state.get(page_state_key, 1))
            current_page = min(max(current_page, 1), total_pages)
            context.session_state[page_state_key] = current_page
            start = (current_page - 1) * self.page_size
            end = start + self.page_size
            frame = frame.iloc[start:end].copy()
            context.st.caption(f"Rows {start + 1}-{min(end, total_rows)} of {total_rows}")
            page = context.st.number_input(
                "Page",
                min_value=1,
                max_value=total_pages,
                value=current_page,
                step=1,
                key=page_widget_key,
            )
            context.session_state[page_state_key] = int(page)
        if self.editable:
            data_editor_kwargs = {
                "key": self.key or self.key_prefix or f"table:{self.order}",
                "width": _container_width_value(self.use_container_width),
                "hide_index": not self.show_index,
            }
            if self.height is not None:
                data_editor_kwargs["height"] = self.height
            rendered = context.st.data_editor(frame, **data_editor_kwargs)
        else:
            dataframe_kwargs = {
                "width": _container_width_value(self.use_container_width),
                "hide_index": not self.show_index,
            }
            if self.height is not None:
                dataframe_kwargs["height"] = self.height
            context.st.dataframe(frame, **dataframe_kwargs)
            rendered = frame
        return rendered


@dataclass(slots=True)
class FileUploadNode(Node):
    accept_multiple_files: bool = False
    type: Optional[Sequence[str]] = None
    _value: Any = field(default=None, init=False, repr=False)

    @property
    def value(self) -> Any:
        if self.key and st is not None and self.key in st.session_state:
            return st.session_state[self.key]
        return self._value

    def _widget_key(self) -> str:
        return self.key or f"file_uploader:{self.order}:{_slugify(self.label)}"

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return self.value
        key = self._widget_key()
        self.key = key
        self._value = context.st.file_uploader(
            self.label,
            key=key,
            accept_multiple_files=self.accept_multiple_files,
            type=self.type,
            help=self.help,
        )
        return self._value


@dataclass(slots=True)
class DownloadButtonNode(Node):
    data: Any = None
    file_name: str = "download"
    mime: Optional[str] = None
    on_click: Optional[Callable[..., Any]] = None
    use_container_width: bool = False

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return False
        payload = _coerce_callable(self.data, context)
        clicked = context.st.download_button(
            self.label,
            data=payload,
            file_name=self.file_name,
            mime=self.mime,
            key=self.key or f"download:{self.order}:{_slugify(self.label)}",
            help=self.help,
            width=_container_width_value(self.use_container_width),
        )
        if clicked:
            _invoke_callback(self.on_click, context)
        return clicked


@dataclass(slots=True)
class InputNode(Node):
    default: Any = ""
    placeholder: Optional[str] = None
    kind: str = "text"
    options: Optional[Sequence[Any]] = None
    min_value: Any = None
    max_value: Any = None
    step: Any = None
    value_type: Optional[str] = None
    _value: Any = field(default=None, init=False, repr=False)

    @property
    def value(self) -> Any:
        if self.key and st is not None and self.key in st.session_state:
            return st.session_state[self.key]
        return self._value if self._value is not None else self.default

    def _widget_key(self) -> str:
        return self.key or f"{self.kind}:{self.order}:{_slugify(self.label)}"

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return self.value
        key = self._widget_key()
        self.key = key
        if self.kind == "textarea":
            rendered = context.st.text_area(
                self.label,
                value=self.default,
                key=key,
                placeholder=self.placeholder,
                help=self.help,
            )
        elif self.kind == "number":
            rendered = context.st.number_input(
                self.label,
                value=self.default if self.default is not None else 0,
                key=key,
                min_value=self.min_value,
                max_value=self.max_value,
                step=self.step,
                help=self.help,
            )
        elif self.kind == "checkbox":
            rendered = context.st.checkbox(
                self.label,
                value=bool(self.default),
                key=key,
                help=self.help,
            )
        elif self.kind == "select":
            rendered = context.st.selectbox(
                self.label,
                options=list(self.options or []),
                index=(
                    0
                    if self.default is None
                    else (
                        max(0, list(self.options or []).index(self.default))
                        if self.options and self.default in self.options
                        else 0
                    )
                ),
                key=key,
                help=self.help,
            )
        else:
            rendered = context.st.text_input(
                self.label,
                value="" if self.default is None else str(self.default),
                key=key,
                placeholder=self.placeholder,
                help=self.help,
            )
        self._value = rendered
        return rendered


@dataclass(slots=True)
class ButtonNode(Node):
    on_click: Optional[Callable[..., Any]] = None
    type: str = "secondary"
    use_container_width: bool = False

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return False
        clicked = context.st.button(
            self.label,
            key=self.key or f"button:{self.order}:{_slugify(self.label)}",
            help=self.help,
            type=self.type,
            width=_container_width_value(self.use_container_width),
        )
        if clicked:
            _invoke_callback(self.on_click, context)
        return clicked


@dataclass(slots=True)
class ContainerNode(Node):
    children: list[Node] = field(default_factory=list)

    def add(self, node: Node) -> Node:
        self.children.append(node)
        return node

    def render_children(self, context: RenderContext) -> None:
        for child in self.children:
            child.render(context)


@dataclass(slots=True)
class SidebarNode(ContainerNode):
    def __enter__(self):
        if self.app is None:
            raise RuntimeError("sidebar is not attached to an app")
        self.app._stack.append(self)
        return self

    def __exit__(self, exc_type, exc, tb):
        if self.app is None:
            return False
        if self.app._stack and self.app._stack[-1] is self:
            self.app._stack.pop()
        return False

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return None
        with context.st.sidebar:
            if self.label:
                context.st.subheader(self.label)
            self.render_children(context)
        return None


@dataclass(slots=True)
class ColumnNode(ContainerNode):
    def __enter__(self):
        if self.app is None:
            raise RuntimeError("column is not attached to an app")
        self.app._stack.append(self)
        return self

    def __exit__(self, exc_type, exc, tb):
        if self.app is None:
            return False
        if self.app._stack and self.app._stack[-1] is self:
            self.app._stack.pop()
        return False


@dataclass(slots=True)
class ColumnsNode(Node):
    specs: Any = 2
    slots: list[ColumnNode] = field(default_factory=list)

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return None
        column_renderers = context.st.columns(self.specs)
        for slot, renderer in zip(self.slots, column_renderers):
            slot_context = RenderContext(app=context.app, st=renderer)
            with renderer:
                slot.render_children(slot_context)
        return None


@dataclass(slots=True)
class TabNode(ContainerNode):
    def __enter__(self):
        if self.app is None:
            raise RuntimeError("tab is not attached to an app")
        self.app._stack.append(self)
        return self

    def __exit__(self, exc_type, exc, tb):
        if self.app is None:
            return False
        if self.app._stack and self.app._stack[-1] is self:
            self.app._stack.pop()
        return False


@dataclass(slots=True)
class TabsNode(Node):
    labels: Sequence[str] = field(default_factory=list)
    slots: list[TabNode] = field(default_factory=list)

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return None
        tab_renderers = context.st.tabs(list(self.labels))
        for slot, renderer in zip(self.slots, tab_renderers):
            slot_context = RenderContext(app=context.app, st=renderer)
            with renderer:
                slot.render_children(slot_context)
        return None


@dataclass(slots=True)
class PageNode(ContainerNode):
    render_fn: Optional[Callable[..., Any]] = None
    icon: Optional[str] = None
    layout: Optional[str] = None

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return None
        if self.label:
            context.st.subheader(self.label)
        if self.render_fn is None:
            for child in self.children:
                child.render(context)
            return None
        try:
            sig = inspect.signature(self.render_fn)
        except (TypeError, ValueError):
            sig = None
        if sig is not None and len(sig.parameters) == 1:
            self.render_fn(context)
        else:
            self.render_fn()
        for child in self.children:
            child.render(context)
        return None


@dataclass(slots=True)
class FormNode(ContainerNode):
    submit_label: str = "Submit"
    on_submit: Optional[Callable[..., Any]] = None

    def __enter__(self):
        if self.app is None:
            raise RuntimeError("form is not attached to an app")
        self.app._stack.append(self)
        return self

    def __exit__(self, exc_type, exc, tb):
        if self.app is None:
            return False
        if self.app._stack and self.app._stack[-1] is self:
            self.app._stack.pop()
        return False

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return None
        if self.label:
            context.st.markdown(f"**{self.label}**")
        with context.st.container(border=True):
            for child in self.children:
                child.render(context)
            submitted = context.st.button(
                self.submit_label,
                key=self.key or f"form:{self.order}:{_slugify(self.label)}",
                width="stretch",
            )
        if submitted:
            _invoke_callback(self.on_submit, context)
        return submitted


@dataclass(slots=True)
class DialogNode(ContainerNode):
    open_label: str = "Open"
    button_label: Optional[str] = None
    expanded: bool = False
    on_open: Optional[Callable[..., Any]] = None

    def __enter__(self):
        if self.app is None:
            raise RuntimeError("dialog is not attached to an app")
        self.app._stack.append(self)
        return self

    def __exit__(self, exc_type, exc, tb):
        if self.app is None:
            return False
        if self.app._stack and self.app._stack[-1] is self:
            self.app._stack.pop()
        return False

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return None
        open_label = self.button_label or self.open_label
        clicked = context.st.button(
            open_label,
            key=self.key or f"dialog:{self.order}:{_slugify(self.label)}",
            width="content",
        )
        if clicked:
            _invoke_callback(self.on_open, context)
        use_popover = hasattr(context.st, "popover")
        if clicked and use_popover:
            with context.st.popover(self.label or open_label):
                for child in self.children:
                    child.render(context)
        elif clicked:
            with context.st.expander(self.label or open_label, expanded=self.expanded):
                for child in self.children:
                    child.render(context)
        return clicked


@dataclass(slots=True)
class SectionNode(ContainerNode):
    level: int = 2

    def __enter__(self):
        if self.app is None:
            raise RuntimeError("section is not attached to an app")
        self.app._stack.append(self)
        return self

    def __exit__(self, exc_type, exc, tb):
        if self.app is None:
            return False
        if self.app._stack and self.app._stack[-1] is self:
            self.app._stack.pop()
        return False

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return None
        if self.level <= 1:
            context.st.title(self.label)
        elif self.level == 2:
            context.st.header(self.label)
        elif self.level == 3:
            context.st.subheader(self.label)
        else:
            context.st.markdown(f"### {self.label}")
        for child in self.children:
            child.render(context)
        return None


@dataclass(slots=True)
class ExpanderNode(ContainerNode):
    expanded: bool = False

    def __enter__(self):
        if self.app is None:
            raise RuntimeError("expander is not attached to an app")
        self.app._stack.append(self)
        return self

    def __exit__(self, exc_type, exc, tb):
        if self.app is None:
            return False
        if self.app._stack and self.app._stack[-1] is self:
            self.app._stack.pop()
        return False

    def render(self, context: RenderContext) -> Any:
        if not self.visible:
            return None
        with context.st.expander(self.label, expanded=self.expanded):
            for child in self.children:
                child.render(context)
        return None


@dataclass(slots=True, init=False)
class App:
    app_title: str
    layout: str
    page_icon: Optional[str]
    _root_nodes: list[Node] = field(default_factory=list, init=False, repr=False)
    _sidebar_nodes: list[SidebarNode] = field(default_factory=list, init=False, repr=False)
    _pages: list[PageNode] = field(default_factory=list, init=False, repr=False)
    _stack: list[ContainerNode] = field(default_factory=list, init=False, repr=False)

    def __init__(
        self,
        title: str = "SoiGia UI",
        layout: str = "wide",
        page_icon: Optional[str] = None,
        *,
        app_title: Optional[str] = None,
    ) -> None:
        self.app_title = title if app_title is None else app_title
        self.layout = layout
        self.page_icon = page_icon
        self._root_nodes = []
        self._sidebar_nodes = []
        self._pages = []
        self._stack = []

    def _register(self, node: Node) -> Node:
        node.app = self
        if self._stack:
            self._stack[-1].add(node)
        else:
            self._root_nodes.append(node)
        return node

    @contextmanager
    def _push(self, container: ContainerNode):
        self._register(container)
        self._stack.append(container)
        try:
            yield container
        finally:
            self._stack.pop()

    def clear(self) -> None:
        global _ORDER
        self._root_nodes.clear()
        self._sidebar_nodes.clear()
        self._pages.clear()
        self._stack.clear()
        _ORDER = itertools.count()

    def text(
        self,
        value: Any,
        *,
        kind: str = "text",
        key: Optional[str] = None,
        help: Optional[str] = None,
        truncate_at: Optional[int] = None,
        expandable: bool = False,
    ) -> TextNode:
        node = TextNode(
            value=value,
            kind=kind,
            key=key,
            help=help,
            truncate_at=truncate_at,
            expandable=expandable,
        )
        return self._register(node)

    def markdown(
        self,
        value: Any,
        *,
        key: Optional[str] = None,
        help: Optional[str] = None,
        truncate_at: Optional[int] = None,
        expandable: bool = False,
    ) -> TextNode:
        node = TextNode(
            value=value, kind="markdown", key=key, help=help, truncate_at=truncate_at, expandable=expandable
        )
        return self._register(node)

    def code(
        self,
        value: Any,
        *,
        key: Optional[str] = None,
        help: Optional[str] = None,
        truncate_at: Optional[int] = None,
        expandable: bool = False,
    ) -> TextNode:
        node = TextNode(value=value, kind="code", key=key, help=help, truncate_at=truncate_at, expandable=expandable)
        return self._register(node)

    def title(self, value: Any, *, key: Optional[str] = None) -> TextNode:
        return self.text(value, kind="title", key=key)

    def header(self, value: Any, *, key: Optional[str] = None) -> TextNode:
        return self.text(value, kind="header", key=key)

    def subheader(self, value: Any, *, key: Optional[str] = None) -> TextNode:
        return self.text(value, kind="subheader", key=key)

    def caption(self, value: Any, *, key: Optional[str] = None) -> TextNode:
        return self.text(value, kind="caption", key=key)

    def json(self, value: Any, *, key: Optional[str] = None, help: Optional[str] = None) -> TextNode:
        node = TextNode(value=value, kind="json", key=key, help=help)
        return self._register(node)

    def success(self, value: Any, *, key: Optional[str] = None, help: Optional[str] = None) -> TextNode:
        node = TextNode(value=value, kind="success", key=key, help=help)
        return self._register(node)

    def info(self, value: Any, *, key: Optional[str] = None, help: Optional[str] = None) -> TextNode:
        node = TextNode(value=value, kind="info", key=key, help=help)
        return self._register(node)

    def warning(self, value: Any, *, key: Optional[str] = None, help: Optional[str] = None) -> TextNode:
        node = TextNode(value=value, kind="warning", key=key, help=help)
        return self._register(node)

    def error(self, value: Any, *, key: Optional[str] = None, help: Optional[str] = None) -> TextNode:
        node = TextNode(value=value, kind="error", key=key, help=help)
        return self._register(node)

    def rerun(self) -> None:
        if st is None:
            return
        st.rerun()

    def stop(self) -> None:
        if st is None:
            return
        st.stop()

    def state_get(self, key: str, default: Any = None) -> Any:
        return st.session_state.get(key, default) if st is not None else default

    def state_set(self, key: str, value: Any) -> Any:
        if st is None:
            return value
        st.session_state[key] = value
        return value

    def state_delete(self, key: str) -> None:
        if st is None:
            return
        if key in st.session_state:
            del st.session_state[key]

    def divider(self, *, key: Optional[str] = None) -> DividerNode:
        return self._register(DividerNode(key=key))

    def metric(
        self,
        label: str,
        value: Any,
        delta: Any = None,
        *,
        delta_color: str = "normal",
        key: Optional[str] = None,
    ) -> MetricNode:
        node = MetricNode(label=label, value=value, delta=delta, delta_color=delta_color, key=key)
        return self._register(node)

    def table(
        self,
        data: Any,
        *,
        key: Optional[str] = None,
        use_container_width: bool = True,
        height: Optional[int] = None,
        editable: bool = False,
        paginate: bool = False,
        page_size: int = 100,
        show_index: bool = False,
        empty_message: str = "No rows to display.",
    ) -> TableNode:
        node = TableNode(
            data=data,
            key=key,
            use_container_width=use_container_width,
            height=height,
            editable=editable,
            paginate=paginate,
            page_size=page_size,
            show_index=show_index,
            empty_message=empty_message,
        )
        return self._register(node)

    def _input(
        self,
        kind: str,
        label: str,
        default: Any = "",
        *,
        key: Optional[str] = None,
        placeholder: Optional[str] = None,
        help: Optional[str] = None,
        options: Optional[Sequence[Any]] = None,
        min_value: Any = None,
        max_value: Any = None,
        step: Any = None,
    ) -> InputNode:
        node = InputNode(
            label=label,
            default=default,
            key=key,
            placeholder=placeholder,
            help=help,
            kind=kind,
            options=options,
            min_value=min_value,
            max_value=max_value,
            step=step,
        )
        return self._register(node)

    def page(
        self,
        title: str,
        *,
        icon: Optional[str] = None,
        layout: Optional[str] = None,
        key: Optional[str] = None,
    ):
        def decorator(fn: Callable[..., Any]) -> Callable[..., Any]:
            node = PageNode(label=title, render_fn=fn, icon=icon, layout=layout, key=key)
            self._pages.append(node)
            return fn

        return decorator

    def text_input(
        self,
        label: str,
        default: Any = "",
        *,
        key: Optional[str] = None,
        placeholder: Optional[str] = None,
        help: Optional[str] = None,
    ) -> InputNode:
        return self._input("text", label, default, key=key, placeholder=placeholder, help=help)

    def text_area(
        self,
        label: str,
        default: Any = "",
        *,
        key: Optional[str] = None,
        placeholder: Optional[str] = None,
        help: Optional[str] = None,
    ) -> InputNode:
        return self._input("textarea", label, default, key=key, placeholder=placeholder, help=help)

    def checkbox(
        self,
        label: str,
        default: bool = False,
        *,
        key: Optional[str] = None,
        help: Optional[str] = None,
    ) -> InputNode:
        return self._input("checkbox", label, default, key=key, help=help)

    def select(
        self,
        label: str,
        options: Sequence[Any],
        default: Any = None,
        *,
        key: Optional[str] = None,
        help: Optional[str] = None,
    ) -> InputNode:
        return self._input("select", label, default, key=key, help=help, options=options)

    def number_input(
        self,
        label: str,
        default: Any = 0,
        *,
        key: Optional[str] = None,
        help: Optional[str] = None,
        min_value: Any = None,
        max_value: Any = None,
        step: Any = None,
    ) -> InputNode:
        return self._input(
            "number",
            label,
            default,
            key=key,
            help=help,
            min_value=min_value,
            max_value=max_value,
            step=step,
        )

    def button(
        self,
        label: str,
        on_click: Optional[Callable[..., Any]] = None,
        *,
        key: Optional[str] = None,
        help: Optional[str] = None,
        type: str = "secondary",
        use_container_width: bool = False,
    ) -> ButtonNode:
        node = ButtonNode(
            label=label,
            on_click=on_click,
            key=key,
            help=help,
            type=type,
            use_container_width=use_container_width,
        )
        return self._register(node)

    def file_uploader(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        help: Optional[str] = None,
        accept_multiple_files: bool = False,
        type: Optional[Sequence[str]] = None,
    ) -> FileUploadNode:
        node = FileUploadNode(
            label=label,
            key=key,
            help=help,
            accept_multiple_files=accept_multiple_files,
            type=type,
        )
        return self._register(node)

    def download_button(
        self,
        label: str,
        data: Any,
        *,
        file_name: str = "download",
        mime: Optional[str] = None,
        on_click: Optional[Callable[..., Any]] = None,
        key: Optional[str] = None,
        help: Optional[str] = None,
        use_container_width: bool = False,
    ) -> DownloadButtonNode:
        node = DownloadButtonNode(
            label=label,
            data=data,
            file_name=file_name,
            mime=mime,
            on_click=on_click,
            key=key,
            help=help,
            use_container_width=use_container_width,
        )
        return self._register(node)

    def form(
        self,
        title: str,
        *,
        submit_label: str = "Submit",
        on_submit: Optional[Callable[..., Any]] = None,
        key: Optional[str] = None,
    ) -> FormNode:
        return self._register(FormNode(label=title, submit_label=submit_label, on_submit=on_submit, key=key))

    def section(self, title: str, *, level: int = 2, key: Optional[str] = None) -> SectionNode:
        return self._register(SectionNode(label=title, level=level, key=key))

    def sidebar(self, title: str = "", *, key: Optional[str] = None) -> SidebarNode:
        node = SidebarNode(label=title, key=key)
        node.app = self
        self._sidebar_nodes.append(node)
        return node

    def columns(self, specs: Any = 2, *, key: Optional[str] = None) -> tuple[ColumnNode, ...]:
        if isinstance(specs, int):
            count = specs
            widths = specs
        else:
            widths = list(specs)
            count = len(widths)
        slots = [ColumnNode(label=f"Column {index + 1}") for index in range(count)]
        node = ColumnsNode(label="Columns", key=key, specs=widths, slots=slots)
        self._register(node)
        for slot in slots:
            slot.app = self
        return tuple(slots)

    def tabs(self, labels: Sequence[str], *, key: Optional[str] = None) -> tuple[TabNode, ...]:
        slots = [TabNode(label=label) for label in labels]
        node = TabsNode(label="Tabs", key=key, labels=list(labels), slots=slots)
        self._register(node)
        for slot in slots:
            slot.app = self
        return tuple(slots)

    @contextmanager
    def section_block(self, title: str, *, level: int = 2, key: Optional[str] = None):
        container = self.section(title, level=level, key=key)
        with container:
            yield container

    @contextmanager
    def expander_block(self, title: str, *, expanded: bool = False, key: Optional[str] = None):
        container = self.expander(title, expanded=expanded, key=key)
        with container:
            yield container

    def expander(self, title: str, *, expanded: bool = False, key: Optional[str] = None) -> ExpanderNode:
        return self._register(ExpanderNode(label=title, expanded=expanded, key=key))

    def dialog(
        self,
        title: str,
        *,
        open_label: str = "Open",
        button_label: Optional[str] = None,
        expanded: bool = False,
        on_open: Optional[Callable[..., Any]] = None,
        key: Optional[str] = None,
    ) -> DialogNode:
        return self._register(
            DialogNode(
                label=title,
                open_label=open_label,
                button_label=button_label,
                expanded=expanded,
                on_open=on_open,
                key=key,
            )
        )

    def render(self) -> None:
        if st is None:
            raise RuntimeError("streamlit is not installed; install it to render the UI")
        try:
            st.set_page_config(page_title=self.app_title, page_icon=self.page_icon, layout=self.layout)
            if self.app_title:
                st.title(self.app_title)
            context = RenderContext(app=self, st=st)
            for node in self._sidebar_nodes:
                node.render(context)
            if self._pages:
                page_labels = [page.label or f"Page {index + 1}" for index, page in enumerate(self._pages)]
                selected_label = st.sidebar.selectbox("Page", page_labels, index=0)
                active_page = self._pages[page_labels.index(selected_label)]
                with self._push(active_page):
                    active_page.render(context)
            else:
                for node in self._root_nodes:
                    node.render(context)
        finally:
            self.clear()

    def run(self) -> None:
        self.render()


# Default singleton, so users can just do `from soigia.ui import ui`.
ui = App()


def run() -> None:
    ui.run()


def rerun() -> None:
    ui.rerun()


def stop() -> None:
    ui.stop()


def state_get(key: str, default: Any = None) -> Any:
    return ui.state_get(key, default)


def state_set(key: str, value: Any) -> Any:
    return ui.state_set(key, value)


def state_delete(key: str) -> None:
    ui.state_delete(key)


def page(
    title: str,
    *,
    icon: Optional[str] = None,
    layout: Optional[str] = None,
    key: Optional[str] = None,
):
    return ui.page(title, icon=icon, layout=layout, key=key)


def text(
    value: Any,
    *,
    key: Optional[str] = None,
    help: Optional[str] = None,
    truncate_at: Optional[int] = None,
    expandable: bool = False,
) -> TextNode:
    return ui.text(value, key=key, help=help, truncate_at=truncate_at, expandable=expandable)


def markdown(
    value: Any,
    *,
    key: Optional[str] = None,
    help: Optional[str] = None,
    truncate_at: Optional[int] = None,
    expandable: bool = False,
) -> TextNode:
    return ui.markdown(value, key=key, help=help, truncate_at=truncate_at, expandable=expandable)


def code(
    value: Any,
    *,
    key: Optional[str] = None,
    help: Optional[str] = None,
    truncate_at: Optional[int] = None,
    expandable: bool = False,
) -> TextNode:
    return ui.code(value, key=key, help=help, truncate_at=truncate_at, expandable=expandable)


def divider(*, key: Optional[str] = None) -> DividerNode:
    return ui.divider(key=key)


def json(value: Any, *, key: Optional[str] = None, help: Optional[str] = None) -> TextNode:
    return ui.json(value, key=key, help=help)


def success(value: Any, *, key: Optional[str] = None, help: Optional[str] = None) -> TextNode:
    return ui.success(value, key=key, help=help)


def info(value: Any, *, key: Optional[str] = None, help: Optional[str] = None) -> TextNode:
    return ui.info(value, key=key, help=help)


def warning(value: Any, *, key: Optional[str] = None, help: Optional[str] = None) -> TextNode:
    return ui.warning(value, key=key, help=help)


def error(value: Any, *, key: Optional[str] = None, help: Optional[str] = None) -> TextNode:
    return ui.error(value, key=key, help=help)


def metric(
    label: str,
    value: Any,
    delta: Any = None,
    *,
    delta_color: str = "normal",
    key: Optional[str] = None,
) -> MetricNode:
    return ui.metric(label, value, delta=delta, delta_color=delta_color, key=key)


def text_input(
    label: str,
    default: Any = "",
    *,
    key: Optional[str] = None,
    placeholder: Optional[str] = None,
    help: Optional[str] = None,
) -> InputNode:
    return ui.text_input(label, default=default, key=key, placeholder=placeholder, help=help)


def textbox(
    label: str,
    default: Any = "",
    *,
    key: Optional[str] = None,
    placeholder: Optional[str] = None,
    help: Optional[str] = None,
) -> InputNode:
    return text_input(label, default=default, key=key, placeholder=placeholder, help=help)


def text_area(
    label: str,
    default: Any = "",
    *,
    key: Optional[str] = None,
    placeholder: Optional[str] = None,
    help: Optional[str] = None,
) -> InputNode:
    return ui.text_area(label, default=default, key=key, placeholder=placeholder, help=help)


def textarea(
    label: str,
    default: Any = "",
    *,
    key: Optional[str] = None,
    placeholder: Optional[str] = None,
    help: Optional[str] = None,
) -> InputNode:
    return text_area(label, default=default, key=key, placeholder=placeholder, help=help)


def checkbox(
    label: str,
    default: bool = False,
    *,
    key: Optional[str] = None,
    help: Optional[str] = None,
) -> InputNode:
    return ui.checkbox(label, default=default, key=key, help=help)


def select(
    label: str,
    options: Sequence[Any],
    default: Any = None,
    *,
    key: Optional[str] = None,
    help: Optional[str] = None,
) -> InputNode:
    return ui.select(label, options, default=default, key=key, help=help)


def button(
    label: str,
    on_click: Optional[Callable[..., Any]] = None,
    *,
    key: Optional[str] = None,
    help: Optional[str] = None,
    type: str = "secondary",
    use_container_width: bool = False,
) -> ButtonNode:
    return ui.button(label, on_click=on_click, key=key, help=help, type=type, use_container_width=use_container_width)


def file_uploader(
    label: str,
    *,
    key: Optional[str] = None,
    help: Optional[str] = None,
    accept_multiple_files: bool = False,
    type: Optional[Sequence[str]] = None,
) -> FileUploadNode:
    return ui.file_uploader(
        label,
        key=key,
        help=help,
        accept_multiple_files=accept_multiple_files,
        type=type,
    )


def download_button(
    label: str,
    data: Any,
    *,
    file_name: str = "download",
    mime: Optional[str] = None,
    on_click: Optional[Callable[..., Any]] = None,
    key: Optional[str] = None,
    help: Optional[str] = None,
    use_container_width: bool = False,
) -> DownloadButtonNode:
    return ui.download_button(
        label,
        data,
        file_name=file_name,
        mime=mime,
        on_click=on_click,
        key=key,
        help=help,
        use_container_width=use_container_width,
    )


def form(
    title: str,
    *,
    submit_label: str = "Submit",
    on_submit: Optional[Callable[..., Any]] = None,
    key: Optional[str] = None,
) -> FormNode:
    return ui.form(title, submit_label=submit_label, on_submit=on_submit, key=key)


def dialog(
    title: str,
    *,
    open_label: str = "Open",
    button_label: Optional[str] = None,
    expanded: bool = False,
    on_open: Optional[Callable[..., Any]] = None,
    key: Optional[str] = None,
) -> DialogNode:
    return ui.dialog(
        title,
        open_label=open_label,
        button_label=button_label,
        expanded=expanded,
        on_open=on_open,
        key=key,
    )


def table(
    data: Any,
    *,
    key: Optional[str] = None,
    use_container_width: bool = True,
    height: Optional[int] = None,
    editable: bool = False,
    paginate: bool = False,
    page_size: int = 100,
    show_index: bool = False,
    empty_message: str = "No rows to display.",
) -> TableNode:
    return ui.table(
        data,
        key=key,
        use_container_width=use_container_width,
        height=height,
        editable=editable,
        paginate=paginate,
        page_size=page_size,
        show_index=show_index,
        empty_message=empty_message,
    )


def sidebar(title: str = "", *, key: Optional[str] = None) -> SidebarNode:
    return ui.sidebar(title, key=key)


def columns(specs: Any = 2, *, key: Optional[str] = None) -> tuple[ColumnNode, ...]:
    return ui.columns(specs, key=key)


def tabs(labels: Sequence[str], *, key: Optional[str] = None) -> tuple[TabNode, ...]:
    return ui.tabs(labels, key=key)


def section(title: str, *, level: int = 2, key: Optional[str] = None) -> SectionNode:
    return ui.section(title, level=level, key=key)


def expander(title: str, *, expanded: bool = False, key: Optional[str] = None) -> ExpanderNode:
    return ui.expander(title, expanded=expanded, key=key)


if __name__ == "__main__":
    run()



"""Scenario-based test cockpit utilities for the SDK."""

from __future__ import annotations

import csv
import json
import os
import re
import shutil
import sqlite3
import subprocess
import sys
import time
import urllib.request
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Mapping, MutableMapping, Optional, Sequence

try:
    import yaml
except Exception:  # pragma: no cover - runtime dependency
    yaml = None  # type: ignore[assignment]


def _require_yaml() -> None:
    if yaml is None:  # pragma: no cover - runtime dependency
        raise RuntimeError("PyYAML is required for Tester YAML helpers")


def _root(root: Optional[Path | str] = None) -> Path:
    return Path(root).expanduser().resolve() if root is not None else Path(__file__).resolve().parents[1]


def _slug(value: str) -> str:
    return re.sub(r"[^A-Za-z0-9._-]+", "_", value.strip()).strip("._-") or "scenario"


def _as_path(value: Path | str, *, root: Optional[Path] = None) -> Path:
    path = Path(value).expanduser()
    return path.resolve() if path.is_absolute() else ((root or Path.cwd()) / path).resolve()


def _norm(text: str) -> str:
    return "\n".join(line.rstrip() for line in text.replace("\r\n", "\n").replace("\r", "\n").split("\n")).strip("\n")


def _read_text(path: Path, *, encoding: str = "utf-8") -> str:
    return path.read_text(encoding=encoding)


def _csv_rows(path: Path, *, delimiter: str = "\t", encoding: str = "utf-8") -> list[list[str]]:
    with path.open("r", encoding=encoding, newline="") as handle:
        return [[cell.strip() for cell in row] for row in csv.reader(handle, delimiter=delimiter)]


def _write_csv(path: Path, rows: Sequence[Sequence[Any]], *, delimiter: str = "\t", encoding: str = "utf-8") -> Path:
    path.parent.mkdir(parents=True, exist_ok=True)
    with path.open("w", encoding=encoding, newline="") as handle:
        writer = csv.writer(handle, delimiter=delimiter)
        for row in rows:
            writer.writerow(list(row))
    return path


class TestMode(str, Enum):
    """Canonical test modes for this SDK."""

    FLOW = "flow"
    STEP = "step"
    SNAPSHOT = "snapshot"
    CONTRACT = "contract"


@dataclass(frozen=True, slots=True)
class TestModeSpec:
    """Description of one test mode."""

    mode: TestMode
    goal: str
    when_to_use: str
    scope: str
    naming_prefix: str
    typical_assertions: tuple[str, ...]

    def as_dict(self) -> dict[str, Any]:
        return {
            "mode": self.mode.value,
            "goal": self.goal,
            "when_to_use": self.when_to_use,
            "scope": self.scope,
            "naming_prefix": self.naming_prefix,
            "typical_assertions": list(self.typical_assertions),
        }


@dataclass(frozen=True, slots=True)
class TestRecipe:
    """Concrete usage recipe for one test mode."""

    mode: TestMode
    title: str
    given: tuple[str, ...]
    when: tuple[str, ...]
    then: tuple[str, ...]
    notes: tuple[str, ...] = ()

    def as_dict(self) -> dict[str, Any]:
        return {
            "mode": self.mode.value,
            "title": self.title,
            "given": list(self.given),
            "when": list(self.when),
            "then": list(self.then),
            "notes": list(self.notes),
        }

    def as_markdown(self) -> str:
        lines = [f"# {self.title}", "", f"- Mode: `{self.mode.value}`", ""]
        for heading, items in (("Given", self.given), ("When", self.when), ("Then", self.then), ("Notes", self.notes)):
            if not items:
                continue
            lines.extend([f"## {heading}", *[f"- {item}" for item in items], ""])
        return "\n".join(lines).rstrip()


@dataclass(frozen=True, slots=True)
class TestPlan:
    """A full test plan: mode + recipe + scenario name."""

    mode: TestMode
    subject: str
    case: str | None = None
    recipe: TestRecipe | None = None

    @property
    def name(self) -> str:
        parts = [self.mode.value, _slug(self.subject)]
        if self.case:
            parts.append(_slug(self.case))
        return "_".join(parts)

    @property
    def title(self) -> str:
        return self.recipe.title if self.recipe is not None else f"{self.mode.value.title()}Test Plan"

    def as_dict(self) -> dict[str, Any]:
        return {
            "mode": self.mode.value,
            "subject": self.subject,
            "case": self.case,
            "name": self.name,
            "title": self.title,
            "recipe": self.recipe.as_dict() if self.recipe is not None else None,
        }

    def as_markdown(self) -> str:
        lines = [f"# {self.title}", "", f"- Mode: `{self.mode.value}`", f"- Subject: `{self.subject}`"]
        if self.case:
            lines.append(f"- Case: `{self.case}`")
        lines.append(f"- Name: `{self.name}`")
        lines.append("")
        if self.recipe is not None:
            lines.append(self.recipe.as_markdown())
        return "\n".join(lines).rstrip()

    def with_recipe(self, recipe: TestRecipe) -> "TestPlan":
        return TestPlan(mode=self.mode, subject=self.subject, case=self.case, recipe=recipe)

    def scenario_name(self) -> str:
        return self.name

    def scenario(self, tester: "Tester") -> "Scenario":
        return tester.scenario(self.scenario_name())

    def pipeline(self, tester: "Tester") -> "PipelineScenario":
        return tester.pipeline(self.scenario_name())

    def run(self, tester: "Tester") -> "Scenario":
        return tester.execute(self)

    def start(self, tester: "Tester") -> "Scenario":
        return self.run(tester)


TEST_MODE_SPECS: tuple[TestModeSpec, ...] = (
    TestModeSpec(
        mode=TestMode.FLOW,
        goal="Test the full business flow end-to-end.",
        when_to_use="Use when you care about the whole orchestration, not an isolated helper.",
        scope="Whole pipeline / scenario / integration flow.",
        naming_prefix="flow",
        typical_assertions=("exit_code", "artifact_exists", "summary_contains", "no_traceback"),
    ),
    TestModeSpec(
        mode=TestMode.STEP,
        goal="Test one step in isolation.",
        when_to_use="Use when you want fast feedback on a single transformation or boundary method.",
        scope="Single step, single boundary, single adapter seam.",
        naming_prefix="step",
        typical_assertions=("row_count", "column_set", "single_output", "call_order"),
    ),
    TestModeSpec(
        mode=TestMode.SNAPSHOT,
        goal="Freeze and compare produced artifacts.",
        when_to_use="Use when the output is stable and you want regression protection on files or tabular output.",
        scope="CSV, SQLite rows, markdown summary, YAML, JSON, logs.",
        naming_prefix="snapshot",
        typical_assertions=("csv_equals", "text_equals", "json_equals", "yaml_equals"),
    ),
    TestModeSpec(
        mode=TestMode.CONTRACT,
        goal="Verify the contract between components or against external boundaries.",
        when_to_use="Use when schema, format, or API compatibility must stay stable.",
        scope="Input/output schema, adapter contract, public API, boundary behavior.",
        naming_prefix="contract",
        typical_assertions=("schema", "types", "required_fields", "allowed_values"),
    ),
)


TEST_MODE_RECIPES: tuple[TestRecipe, ...] = (
    TestRecipe(
        mode=TestMode.FLOW,
        title="FlowTest Recipe",
        given=(
            "Prepare the full input dataset.",
            "Seed config.yaml or fixtures required by the business flow.",
            "Seed golden files only if you want to compare stable artifacts.",
        ),
        when=(
            "Run the pipeline, command, or end-to-end scenario.",
        ),
        then=(
            "Assert exit code is zero.",
            "Assert no traceback occurred.",
            "Assert pipeline summary and final artifacts exist.",
            "Assert final rows, columns, and log markers if needed.",
        ),
        notes=(
            "Use this when you care about the whole orchestration.",
            "Prefer this for the public business flow path.",
        ),
    ),
    TestRecipe(
        mode=TestMode.STEP,
        title="StepTest Recipe",
        given=(
            "Prepare the smallest possible input for one step.",
            "Mock only the external boundary touched by that step.",
        ),
        when=(
            "Run only the selected step or call the seam method directly.",
        ),
        then=(
            "Assert the step output shape, values, and side effects.",
            "Assert call order if the step interacts with a fake boundary.",
        ),
        notes=(
            "Use this for fast feedback on a single transformation.",
            "Keep it narrow; do not load the whole flow unless necessary.",
        ),
    ),
    TestRecipe(
        mode=TestMode.SNAPSHOT,
        title="SnapshotTest Recipe",
        given=(
            "Prepare a known-good baseline artifact or golden file.",
            "Stabilize non-deterministic fields before comparison.",
        ),
        when=(
            "Run the code that emits CSV, markdown, JSON, YAML, SQLite rows, or logs.",
        ),
        then=(
            "Assert the produced artifact matches the golden snapshot.",
            "Use normalized text/CSV comparison when whitespace may vary.",
        ),
        notes=(
            "Use this for regression protection on outputs.",
            "Prefer this for pipeline step dumps and summary reports.",
        ),
    ),
    TestRecipe(
        mode=TestMode.CONTRACT,
        title="ContractTest Recipe",
        given=(
            "Prepare a contract description or schema expectation.",
            "Define the required columns, types, keys, or API behavior.",
        ),
        when=(
            "Run the adapter, boundary, or public API that must obey the contract.",
        ),
        then=(
            "Assert schema, required fields, allowed values, and types.",
            "Assert boundary compatibility rather than full content equality.",
        ),
        notes=(
            "Use this to protect interoperability with external systems.",
            "This is the right mode for Google Sheet, Binance, MT5, Telegram, DB, and HTTP boundaries.",
        ),
    ),
)


def _mode_recipe_map() -> dict[TestMode, TestRecipe]:
    return {recipe.mode: recipe for recipe in TEST_MODE_RECIPES}


def _mode_spec_map() -> dict[TestMode, TestModeSpec]:
    return {spec.mode: spec for spec in TEST_MODE_SPECS}


def _mode_from_string(value: str) -> TestMode | None:
    normalized = value.strip().lower()
    for mode in TestMode:
        if normalized == mode.value:
            return mode
    return None


@dataclass(slots=True)
class CommandResult:
    __test__ = False

    command: list[str]
    cwd: Path
    returncode: int
    stdout: str
    stderr: str
    duration_seconds: float
    started_at: float
    finished_at: float

    @property
    def success(self) -> bool:
        return self.returncode == 0

    def check(self) -> "CommandResult":
        if not self.success:
            raise AssertionError(self.format_failure())
        return self

    def format_failure(self, preview: int = 4000) -> str:
        return (
            f"Command failed: {' '.join(self.command)}\n"
            f"cwd: {self.cwd}\n"
            f"returncode: {self.returncode}\n"
            f"stdout:\n{self.stdout[-preview:]}\n"
            f"stderr:\n{self.stderr[-preview:]}"
        )

    def as_dict(self) -> dict[str, Any]:
        return {
            "command": self.command,
            "cwd": str(self.cwd),
            "returncode": self.returncode,
            "stdout": self.stdout,
            "stderr": self.stderr,
            "duration_seconds": self.duration_seconds,
            "started_at": self.started_at,
            "finished_at": self.finished_at,
            "success": self.success,
        }


@dataclass(slots=True)
class Tester:
    """Private-project friendly helper for scenario-style tests."""

    __test__ = False

    root: Path | str | None = None
    artifacts_dir_name: str = ".tester_artifacts"
    golden_dir_name: str = "golden"

    def root_dir(self) -> Path:
        return _root(self.root)

    def artifacts_dir(self) -> Path:
        path = self.root_dir() / self.artifacts_dir_name
        path.mkdir(parents=True, exist_ok=True)
        return path

    def golden_dir(self) -> Path:
        return self.root_dir() / self.golden_dir_name

    def path(self, *parts: str) -> Path:
        return self.artifacts_dir().joinpath(*parts)

    def mode_spec(self, mode: TestMode | str) -> TestModeSpec:
        resolved = _mode_from_string(mode) if isinstance(mode, str) else mode
        if resolved is None:
            raise ValueError(f"Unknown test mode: {mode!r}")
        return _mode_spec_map()[resolved]

    def mode_specs(self) -> tuple[TestModeSpec, ...]:
        return TEST_MODE_SPECS

    def recipe(self, mode: TestMode | str) -> TestRecipe:
        resolved = _mode_from_string(mode) if isinstance(mode, str) else mode
        if resolved is None:
            raise ValueError(f"Unknown test mode: {mode!r}")
        return _mode_recipe_map()[resolved]

    def flow_recipe(self) -> TestRecipe:
        return self.recipe(TestMode.FLOW)

    def step_recipe(self) -> TestRecipe:
        return self.recipe(TestMode.STEP)

    def snapshot_recipe(self) -> TestRecipe:
        return self.recipe(TestMode.SNAPSHOT)

    def contract_recipe(self) -> TestRecipe:
        return self.recipe(TestMode.CONTRACT)

    def plan(self, mode: TestMode | str, subject: str, case: str | None = None) -> TestPlan:
        resolved = _mode_from_string(mode) if isinstance(mode, str) else mode
        if resolved is None:
            raise ValueError(f"Unknown test mode: {mode!r}")
        return TestPlan(mode=resolved, subject=subject, case=case, recipe=self.recipe(resolved))

    def flow_plan(self, subject: str, case: str | None = None) -> TestPlan:
        return self.plan(TestMode.FLOW, subject, case)

    def step_plan(self, subject: str, case: str | None = None) -> TestPlan:
        return self.plan(TestMode.STEP, subject, case)

    def snapshot_plan(self, subject: str, case: str | None = None) -> TestPlan:
        return self.plan(TestMode.SNAPSHOT, subject, case)

    def contract_plan(self, subject: str, case: str | None = None) -> TestPlan:
        return self.plan(TestMode.CONTRACT, subject, case)

    def flow_case(self, subject: str, case: str) -> TestPlan:
        return self.flow_plan(subject, case)

    def step_case(self, subject: str, case: str) -> TestPlan:
        return self.step_plan(subject, case)

    def snapshot_case(self, subject: str, case: str) -> TestPlan:
        return self.snapshot_plan(subject, case)

    def contract_case(self, subject: str, case: str) -> TestPlan:
        return self.contract_plan(subject, case)

    def boundary_contract(self, boundary: str, subject: str, case: str | None = None) -> TestPlan:
        """Shortcut for boundary-focused contract tests."""

        return self.contract_plan(f"{_slug(boundary)}_{_slug(subject)}", case)

    def sheet_contract(self, subject: str, case: str | None = None) -> TestPlan:
        return self.boundary_contract("sheet", subject, case)

    def binance_contract(self, subject: str, case: str | None = None) -> TestPlan:
        return self.boundary_contract("binance", subject, case)

    def mt5_contract(self, subject: str, case: str | None = None) -> TestPlan:
        return self.boundary_contract("mt5", subject, case)

    def telegram_contract(self, subject: str, case: str | None = None) -> TestPlan:
        return self.boundary_contract("telegram", subject, case)

    def database_contract(self, subject: str, case: str | None = None) -> TestPlan:
        return self.boundary_contract("database", subject, case)

    def http_contract(self, subject: str, case: str | None = None) -> TestPlan:
        return self.boundary_contract("http", subject, case)

    def execute(self, plan: TestPlan) -> "Scenario":
        """Instantiate the correct scenario type for a plan."""

        if plan.mode in {TestMode.FLOW}:
            return plan.pipeline(self)
        return plan.scenario(self)

    def start(self, plan: TestPlan) -> "Scenario":
        return self.execute(plan)

    def run(self, plan: TestPlan) -> "Scenario":
        return self.execute(plan)

    def recipes(self) -> tuple[TestRecipe, ...]:
        return TEST_MODE_RECIPES

    def taxonomy(self) -> dict[str, Any]:
        """Return the 4-mode test taxonomy as serializable data."""

        return {
            "modes": [spec.as_dict() for spec in TEST_MODE_SPECS],
            "naming_convention": self.naming_convention(),
        }

    def mode_guide(self) -> str:
        """Render the test taxonomy as Markdown."""

        lines = ["# Soigia Test Taxonomy", ""]
        for spec in TEST_MODE_SPECS:
            lines.extend(
                [
                    f"## {spec.mode.value.title()}Test",
                    f"- Goal: {spec.goal}",
                    f"- When to use: {spec.when_to_use}",
                    f"- Scope: {spec.scope}",
                    f"- Naming prefix: `{spec.naming_prefix}`",
                    f"- Typical assertions: {', '.join(spec.typical_assertions)}",
                    "",
                ]
            )
        lines.append("## Naming Convention")
        lines.extend(
            [
                f"- Test file: `{self.naming_convention()['file_pattern']}`",
                f"- Test function: `{self.naming_convention()['function_pattern']}`",
                f"- Scenario folder: `{self.naming_convention()['scenario_pattern']}`",
                f"- Artifact folder: `{self.naming_convention()['artifact_pattern']}`",
            ]
        )
        return "\n".join(lines)

    def examples(self) -> dict[str, str]:
        """Return canonical usage examples for the 4 test modes."""

        return {
            "flow": (
                "tester.flow('sales_flow')\n"
                "    .given_yaml('config.yaml', {'source': 'sample'})\n"
                "    .given_csv('input/orders.tsv', [['order_id', 'amount'], ['A1', 100]])\n"
                "    .when_run_module('examples.pipeline_yaml_config_demo')\n"
                "    .then_exit_code(0)\n"
                "    .then_no_traceback()\n"
                "    .then_pipeline_summary_exists()\n"
            ),
            "step": (
                "tester.step('normalize_step')\n"
                "    .given_csv('input.csv', [['order_id', 'amount'], ['A1', 100]])\n"
                "    .when_run_command(['python', '-c', 'print(\"normalize\")'])\n"
                "    .then_exit_code(0)\n"
                "    .then_stdout_contains('normalize')\n"
            ),
            "snapshot": (
                "tester.snapshot('summary_snapshot')\n"
                "    .given_text('expected.md', '# Report\\n')\n"
                "    .when_run_command(['python', '-c', 'print(\"# Report\")'])\n"
                "    .then_exit_code(0)\n"
                "    .then_stdout_contains('Report')\n"
            ),
            "contract": (
                "tester.contract('sheet_contract')\n"
                "    .given_yaml('contract.yaml', {'required_columns': ['id', 'amount']})\n"
                "    .when_run_command(['python', '-c', 'print(\"contract ok\")'])\n"
                "    .then_exit_code(0)\n"
                "    .then_stdout_contains('contract ok')\n"
            ),
        }

    def examples_markdown(self) -> str:
        """Render the canonical usage examples as Markdown."""

        lines = ["# Soigia Tester Examples", ""]
        for mode, example in self.examples().items():
            lines.extend([f"## {mode.title()}Test", "```python", example.rstrip(), "```", ""])
        return "\n".join(lines)

    def recipe_markdown(self) -> str:
        """Render the canonical recipes as Markdown."""

        lines = ["# Soigia Test Recipes", ""]
        for recipe in TEST_MODE_RECIPES:
            lines.append(recipe.as_markdown())
            lines.append("")
        return "\n".join(lines).rstrip()

    def recipe_guide(self) -> dict[str, Any]:
        """Return recipe guidance as serializable data."""

        return {
            "recipes": [recipe.as_dict() for recipe in TEST_MODE_RECIPES],
            "recipes_markdown": self.recipe_markdown(),
        }

    def naming_convention(self) -> dict[str, str]:
        """Return the naming convention used by the four test modes."""

        return {
            "file_pattern": "tests/test_<mode>_<subject>.py",
            "function_pattern": "test_<mode>_<subject>_<case>",
            "scenario_pattern": ".tester_artifacts/<mode>/<subject>/<case>",
            "artifact_pattern": "data/<pipeline_name>/* or logs/<pipeline_name>.log",
        }

    def test_name(self, mode: TestMode | str, subject: str, case: str | None = None) -> str:
        """Build a canonical test name for the selected mode."""

        spec = self.mode_spec(mode)
        parts = [f"test_{spec.naming_prefix}", _slug(subject)]
        if case:
            parts.append(_slug(case))
        return "_".join(parts)

    def recommend_mode(self, intent: str) -> TestMode:
        """Best-effort mode recommendation from a short intent string."""

        text = intent.lower()
        if any(keyword in text for keyword in ("snapshot", "golden", "artifact", "csv", "parquet", "markdown", "yaml", "json")):
            return TestMode.SNAPSHOT
        if any(keyword in text for keyword in ("contract", "schema", "boundary", "adapter", "api", "compat")):
            return TestMode.CONTRACT
        if any(keyword in text for keyword in ("step", "single", "isolate", "isolated", "unit")):
            return TestMode.STEP
        return TestMode.FLOW

    def scenario(self, name: str = "scenario") -> "Scenario":
        return Scenario(self, name=name)

    def pipeline(self, name: str = "pipeline") -> "PipelineScenario":
        return PipelineScenario(self, name=name)

    def flow(self, name: str = "flow") -> "FlowScenario":
        return FlowScenario(self, name=name)

    def step(self, name: str = "step") -> "StepScenario":
        return StepScenario(self, name=name)

    def snapshot(self, name: str = "snapshot") -> "SnapshotScenario":
        return SnapshotScenario(self, name=name)

    def contract(self, name: str = "contract") -> "ContractScenario":
        return ContractScenario(self, name=name)

    def write_text(self, relative_path: str, content: str, *, encoding: str = "utf-8") -> Path:
        path = self.path(relative_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(content, encoding=encoding)
        return path

    def read_text(self, relative_path: str, *, encoding: str = "utf-8") -> str:
        return self.path(relative_path).read_text(encoding=encoding)

    def write_json(self, relative_path: str, data: Any, *, indent: int = 2, sort_keys: bool = True) -> Path:
        path = self.path(relative_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w", encoding="utf-8") as handle:
            json.dump(data, handle, ensure_ascii=False, indent=indent, sort_keys=sort_keys)
            handle.write("\n")
        return path

    def read_json(self, relative_path: str) -> Any:
        with self.path(relative_path).open("r", encoding="utf-8") as handle:
            return json.load(handle)

    def write_yaml(self, relative_path: str, data: Mapping[str, Any], *, sort_keys: bool = False) -> Path:
        _require_yaml()
        path = self.path(relative_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w", encoding="utf-8") as handle:
            yaml.safe_dump(data, handle, sort_keys=sort_keys, allow_unicode=True)
        return path

    def read_yaml(self, relative_path: str) -> Any:
        _require_yaml()
        with self.path(relative_path).open("r", encoding="utf-8") as handle:
            return yaml.safe_load(handle)

    def write_csv(self, relative_path: str, rows: Sequence[Sequence[Any]], *, delimiter: str = "\t") -> Path:
        return _write_csv(self.path(relative_path), rows, delimiter=delimiter)

    def read_csv(self, relative_path: str, *, delimiter: str = "\t") -> list[list[str]]:
        return _csv_rows(self.path(relative_path), delimiter=delimiter)

    def write_sqlite_table(self, relative_path: str, rows: Sequence[Mapping[str, Any]], *, table_name: str = "data") -> Path:
        path = self.path(relative_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        rows = list(rows)
        with sqlite3.connect(path) as connection:
            if not rows:
                connection.execute(f"CREATE TABLE IF NOT EXISTS {table_name} (id INTEGER)")
                connection.commit()
                return path
            columns = list(rows[0].keys())
            connection.execute(
                f'CREATE TABLE IF NOT EXISTS {table_name} ({", ".join(f"""\"{c}\" TEXT""" for c in columns)})'
            )
            connection.execute(f"DELETE FROM {table_name}")
            placeholders = ", ".join("?" for _ in columns)
            insert_sql = f'INSERT INTO {table_name} ({", ".join(columns)}) VALUES ({placeholders})'
            for row in rows:
                connection.execute(insert_sql, [row.get(column) for column in columns])
            connection.commit()
        return path

    def read_sqlite_table(self, relative_path: str, *, table_name: str = "data") -> list[dict[str, Any]]:
        with sqlite3.connect(self.path(relative_path)) as connection:
            connection.row_factory = sqlite3.Row
            return [dict(row) for row in connection.execute(f"SELECT * FROM {table_name}").fetchall()]

    def sqlite_row_count(self, relative_path: str, *, table_name: str = "data") -> int:
        with sqlite3.connect(self.path(relative_path)) as connection:
            row = connection.execute(f"SELECT COUNT(*) FROM {table_name}").fetchone()
            return int(row[0]) if row else 0

    def exists(self, relative_path: str) -> bool:
        return self.path(relative_path).exists()

    def copy_file(self, source: Path | str, relative_target: str) -> Path:
        src = _as_path(source, root=self.root_dir())
        dst = self.path(relative_target)
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(src, dst)
        return dst

    def assert_file_exists(self, relative_path: str) -> Path:
        path = self.path(relative_path)
        if not path.exists():
            raise AssertionError(f"Expected file to exist: {path}")
        return path

    def assert_text_contains(self, relative_path: str, needle: str, *, encoding: str = "utf-8") -> Path:
        path = self.assert_file_exists(relative_path)
        if needle not in _read_text(path, encoding=encoding):
            raise AssertionError(f"{needle!r} not found in {path}")
        return path

    def assert_text_equals(self, relative_path: str, expected: str, *, encoding: str = "utf-8") -> Path:
        path = self.assert_file_exists(relative_path)
        if _norm(_read_text(path, encoding=encoding)) != _norm(expected):
            raise AssertionError(f"Text mismatch in {path}")
        return path

    def assert_json_equals(self, relative_path: str, expected: Any) -> Path:
        path = self.assert_file_exists(relative_path)
        actual = self.read_json(relative_path)
        if actual != expected:
            raise AssertionError(f"JSON mismatch in {path}\nactual={actual!r}\nexpected={expected!r}")
        return path

    def assert_yaml_equals(self, relative_path: str, expected: Any) -> Path:
        path = self.assert_file_exists(relative_path)
        actual = self.read_yaml(relative_path)
        if actual != expected:
            raise AssertionError(f"YAML mismatch in {path}\nactual={actual!r}\nexpected={expected!r}")
        return path

    def assert_csv_equals(self, relative_path: str, expected_rows: Sequence[Sequence[Any]], *, delimiter: str = "\t") -> Path:
        path = self.assert_file_exists(relative_path)
        actual = _csv_rows(path, delimiter=delimiter)
        expected = [[str(cell).strip() for cell in row] for row in expected_rows]
        if actual != expected:
            raise AssertionError(f"CSV mismatch in {path}\nactual={actual!r}\nexpected={expected!r}")
        return path

    def assert_sqlite_row_count(self, relative_path: str, expected_count: int, *, table_name: str = "data") -> Path:
        path = self.assert_file_exists(relative_path)
        actual_count = self.sqlite_row_count(relative_path, table_name=table_name)
        if actual_count != expected_count:
            raise AssertionError(f"Row count mismatch in {path}: {actual_count} != {expected_count}")
        return path

    def assert_sqlite_columns(self, relative_path: str, expected_columns: Sequence[str], *, table_name: str = "data") -> Path:
        path = self.assert_file_exists(relative_path)
        with sqlite3.connect(path) as connection:
            rows = connection.execute(f"PRAGMA table_info({table_name})").fetchall()
        actual = [str(row[1]) for row in rows]
        expected = list(expected_columns)
        if actual != expected:
            raise AssertionError(f"Column mismatch in {path}: {actual!r} != {expected!r}")
        return path

    def run_command(
        self,
        command: Sequence[str] | str,
        *,
        cwd: Path | str | None = None,
        env: Optional[Mapping[str, str]] = None,
        timeout: float | None = None,
        check: bool = False,
    ) -> CommandResult:
        args = command if isinstance(command, str) else [str(item) for item in command]
        working_dir = _as_path(cwd, root=self.root_dir()) if cwd is not None else self.root_dir()
        runtime_env: MutableMapping[str, str] = dict(os.environ)
        if env:
            runtime_env.update({str(k): str(v) for k, v in env.items()})
        started_at = time.time()
        start = time.perf_counter()
        completed = subprocess.run(
            args,
            cwd=str(working_dir),
            env=runtime_env,
            capture_output=True,
            text=True,
            timeout=timeout,
            shell=isinstance(command, str),
        )
        result = CommandResult(
            command=args if isinstance(args, list) else [args],
            cwd=working_dir,
            returncode=completed.returncode,
            stdout=completed.stdout,
            stderr=completed.stderr,
            duration_seconds=time.perf_counter() - start,
            started_at=started_at,
            finished_at=time.time(),
        )
        if check:
            result.check()
        return result

    def run_module(self, module: str, *, args: Sequence[str] = (), cwd: Path | str | None = None, env: Optional[Mapping[str, str]] = None, timeout: float | None = None, check: bool = False) -> CommandResult:
        return self.run_command([sys.executable, "-m", module, *[str(arg) for arg in args]], cwd=cwd, env=env, timeout=timeout, check=check)

    def run_script(self, script: Path | str, *, args: Sequence[str] = (), cwd: Path | str | None = None, env: Optional[Mapping[str, str]] = None, timeout: float | None = None, check: bool = False) -> CommandResult:
        return self.run_command([sys.executable, str(_as_path(script, root=self.root_dir())), *[str(arg) for arg in args]], cwd=cwd, env=env, timeout=timeout, check=check)

    def cleanup(self) -> None:
        shutil.rmtree(self.artifacts_dir(), ignore_errors=True)


@dataclass(slots=True)
class Scenario:
    """Given / When / Then runner for end-to-end cases."""

    __test__ = False

    tester: Tester
    name: str = "scenario"
    result: CommandResult | None = None

    @property
    def scenario_path(self) -> Path:
        return self.tester.artifacts_dir() / _slug(self.name)

    def path(self, *parts: str) -> Path:
        return self.scenario_path.joinpath(*parts)

    def given_text(self, relative_path: str, content: str, *, encoding: str = "utf-8") -> "Scenario":
        path = self.path(relative_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(content, encoding=encoding)
        return self

    def given_json(self, relative_path: str, data: Any) -> "Scenario":
        path = self.path(relative_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(json.dumps(data, ensure_ascii=False, indent=2) + "\n", encoding="utf-8")
        return self

    def given_yaml(self, relative_path: str, data: Mapping[str, Any], *, sort_keys: bool = False) -> "Scenario":
        _require_yaml()
        path = self.path(relative_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w", encoding="utf-8") as handle:
            yaml.safe_dump(data, handle, sort_keys=sort_keys, allow_unicode=True)
        return self

    def given_csv(self, relative_path: str, rows: Sequence[Sequence[Any]], *, delimiter: str = "\t") -> "Scenario":
        _write_csv(self.path(relative_path), rows, delimiter=delimiter)
        return self

    def given_file(self, source: Path | str, relative_target: str) -> "Scenario":
        src = _as_path(source, root=self.tester.root_dir())
        dst = self.path(relative_target)
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(src, dst)
        return self

    def when_run_command(self, command: Sequence[str] | str, *, cwd: Path | str | None = None, env: Optional[Mapping[str, str]] = None, timeout: float | None = None, check: bool = False) -> "Scenario":
        self.result = self.tester.run_command(command, cwd=cwd or self.scenario_path, env=env, timeout=timeout, check=check)
        return self

    def when_run_module(self, module: str, *, args: Sequence[str] = (), cwd: Path | str | None = None, env: Optional[Mapping[str, str]] = None, timeout: float | None = None, check: bool = False) -> "Scenario":
        self.result = self.tester.run_module(module, args=args, cwd=cwd or self.scenario_path, env=env, timeout=timeout, check=check)
        return self

    def when_run_script(self, script: Path | str, *, args: Sequence[str] = (), cwd: Path | str | None = None, env: Optional[Mapping[str, str]] = None, timeout: float | None = None, check: bool = False) -> "Scenario":
        self.result = self.tester.run_script(script, args=args, cwd=cwd or self.scenario_path, env=env, timeout=timeout, check=check)
        return self

    def then_exit_code(self, expected: int = 0) -> "Scenario":
        self._require_result()
        if self.result.returncode != expected:
            raise AssertionError(self.result.format_failure())
        return self

    def then_no_traceback(self) -> "Scenario":
        self._require_result()
        content = f"{self.result.stdout}\n{self.result.stderr}"
        if "Traceback (most recent call last)" in content or "Traceback:" in content:
            raise AssertionError(self.result.format_failure())
        return self

    def then_stdout_contains(self, needle: str) -> "Scenario":
        self._require_result()
        if needle not in self.result.stdout:
            raise AssertionError(f"{needle!r} not found in stdout")
        return self

    def then_stderr_contains(self, needle: str) -> "Scenario":
        self._require_result()
        if needle not in self.result.stderr:
            raise AssertionError(f"{needle!r} not found in stderr")
        return self

    def then_file_exists(self, relative_path: str) -> "Scenario":
        if not self.path(relative_path).exists():
            raise AssertionError(f"Expected file to exist: {self.path(relative_path)}")
        return self

    def then_text_contains(self, relative_path: str, needle: str, *, encoding: str = "utf-8") -> "Scenario":
        if needle not in _read_text(self.path(relative_path), encoding=encoding):
            raise AssertionError(f"{needle!r} not found in {self.path(relative_path)}")
        return self

    def then_text_equals(self, relative_path: str, expected: str, *, encoding: str = "utf-8") -> "Scenario":
        if _norm(_read_text(self.path(relative_path), encoding=encoding)) != _norm(expected):
            raise AssertionError(f"Text mismatch in {self.path(relative_path)}")
        return self

    def then_json_equals(self, relative_path: str, expected: Any) -> "Scenario":
        with self.path(relative_path).open("r", encoding="utf-8") as handle:
            actual = json.load(handle)
        if actual != expected:
            raise AssertionError(f"JSON mismatch in {self.path(relative_path)}\nactual={actual!r}\nexpected={expected!r}")
        return self

    def then_yaml_equals(self, relative_path: str, expected: Any) -> "Scenario":
        _require_yaml()
        with self.path(relative_path).open("r", encoding="utf-8") as handle:
            actual = yaml.safe_load(handle)
        if actual != expected:
            raise AssertionError(f"YAML mismatch in {self.path(relative_path)}\nactual={actual!r}\nexpected={expected!r}")
        return self

    def then_csv_equals(self, relative_path: str, expected_rows: Sequence[Sequence[Any]], *, delimiter: str = "\t") -> "Scenario":
        actual = _csv_rows(self.path(relative_path), delimiter=delimiter)
        expected = [[str(cell).strip() for cell in row] for row in expected_rows]
        if actual != expected:
            raise AssertionError(f"CSV mismatch in {self.path(relative_path)}\nactual={actual!r}\nexpected={expected!r}")
        return self

    def then_sqlite_row_count(self, relative_path: str, expected_count: int, *, table_name: str = "data") -> "Scenario":
        actual = self.tester.sqlite_row_count(str(self.path(relative_path).relative_to(self.tester.artifacts_dir())), table_name=table_name)
        if actual != expected_count:
            raise AssertionError(f"Row count mismatch in {self.path(relative_path)}: {actual} != {expected_count}")
        return self

    def dump_debug(self) -> Path:
        dump_dir = self.scenario_path / "debug"
        dump_dir.mkdir(parents=True, exist_ok=True)
        if self.result is not None:
            (dump_dir / "command.txt").write_text(" ".join(self.result.command), encoding="utf-8")
            (dump_dir / "stdout.txt").write_text(self.result.stdout, encoding="utf-8")
            (dump_dir / "stderr.txt").write_text(self.result.stderr, encoding="utf-8")
            (dump_dir / "result.json").write_text(json.dumps(self.result.as_dict(), ensure_ascii=False, indent=2), encoding="utf-8")
        return dump_dir

    def _require_result(self) -> None:
        if self.result is None:
            raise AssertionError("Scenario has no command result yet. Call when_run_* first.")


@dataclass(slots=True)
class PipelineScenario(Scenario):
    """Scenario runner with pipeline-specific assertions."""

    __test__ = False

    @property
    def pipeline_name(self) -> str:
        return self.name

    def pipeline_dir(self) -> Path:
        return self.tester.root_dir() / "data" / _slug(self.pipeline_name)

    def pipeline_summary_path(self) -> Path:
        return self.pipeline_dir() / "pipeline_summary.md"

    def pipeline_log_path(self) -> Path:
        return self.tester.root_dir() / "logs" / f"{_slug(self.pipeline_name)}.log"

    def then_pipeline_summary_exists(self) -> "PipelineScenario":
        path = self.pipeline_summary_path()
        if not path.exists():
            raise AssertionError(f"Missing pipeline summary: {path}")
        return self

    def then_pipeline_summary_contains(self, needle: str) -> "PipelineScenario":
        path = self.pipeline_summary_path()
        if not path.exists():
            raise AssertionError(f"Missing pipeline summary: {path}")
        if needle not in _read_text(path, encoding="utf-8"):
            raise AssertionError(f"{needle!r} not found in pipeline summary {path}")
        return self

    def then_pipeline_log_contains(self, needle: str) -> "PipelineScenario":
        path = self.pipeline_log_path()
        if not path.exists():
            raise AssertionError(f"Missing pipeline log: {path}")
        if needle not in _read_text(path, encoding="utf-8"):
            raise AssertionError(f"{needle!r} not found in pipeline log {path}")
        return self

    def then_pipeline_artifact_exists(self, relative_path: str) -> "PipelineScenario":
        path = self.pipeline_dir() / relative_path
        if not path.exists():
            raise AssertionError(f"Expected pipeline artifact to exist: {path}")
        return self

    def then_pipeline_csv_equals(
        self,
        relative_path: str,
        expected_rows: Sequence[Sequence[Any]],
        *,
        delimiter: str = "\t",
    ) -> "PipelineScenario":
        path = self.pipeline_dir() / relative_path
        actual = _csv_rows(path, delimiter=delimiter)
        expected = [[str(cell).strip() for cell in row] for row in expected_rows]
        if actual != expected:
            raise AssertionError(f"CSV mismatch in {path}\nactual={actual!r}\nexpected={expected!r}")
        return self

    def then_pipeline_sqlite_row_count(self, relative_path: str, expected_count: int, *, table_name: str = "data") -> "PipelineScenario":
        path = self.pipeline_dir() / relative_path
        with sqlite3.connect(path) as connection:
            row = connection.execute(f"SELECT COUNT(*) FROM {table_name}").fetchone()
            actual = int(row[0]) if row else 0
        if actual != expected_count:
            raise AssertionError(f"Row count mismatch in {path}: {actual} != {expected_count}")
        return self


@dataclass(slots=True)
class FlowScenario(PipelineScenario):
    __test__ = False

    mode: TestMode = TestMode.FLOW


@dataclass(slots=True)
class StepScenario(Scenario):
    __test__ = False

    mode: TestMode = TestMode.STEP


@dataclass(slots=True)
class SnapshotScenario(Scenario):
    __test__ = False

    mode: TestMode = TestMode.SNAPSHOT


@dataclass(slots=True)
class ContractScenario(Scenario):
    __test__ = False

    mode: TestMode = TestMode.CONTRACT


@dataclass(slots=True)
class PipelineTester(Tester):
    __test__ = False

    def pipeline_dir(self, pipeline_name: str) -> Path:
        return self.root_dir() / "data" / _slug(pipeline_name)

    def pipeline_summary_path(self, pipeline_name: str) -> Path:
        return self.pipeline_dir(pipeline_name) / "pipeline_summary.md"

    def pipeline_log_path(self, pipeline_name: str) -> Path:
        return self.root_dir() / "logs" / f"{_slug(pipeline_name)}.log"


@dataclass(slots=True)
class UITester(Tester):
    __test__ = False

    def run_streamlit_script(
        self,
        script: Path | str,
        *,
        port: int = 8501,
        timeout: float = 60.0,
        args: Sequence[str] = (),
        cwd: Path | str | None = None,
        env: Optional[Mapping[str, str]] = None,
    ) -> CommandResult:
        script_path = _as_path(script, root=self.root_dir())
        command = [
            sys.executable,
            "-m",
            "streamlit",
            "run",
            str(script_path),
            "--server.headless",
            "true",
            "--server.address",
            "127.0.0.1",
            "--server.port",
            str(port),
            "--logger.level",
            "error",
            *[str(arg) for arg in args],
        ]
        working_dir = _as_path(cwd, root=self.root_dir()) if cwd is not None else self.root_dir()
        runtime_env: MutableMapping[str, str] = dict(os.environ)
        runtime_env.setdefault("STREAMLIT_BROWSER_GATHER_USAGE_STATS", "false")
        if env:
            runtime_env.update({str(k): str(v) for k, v in env.items()})

        started_at = time.time()
        start = time.perf_counter()
        proc = subprocess.Popen(command, cwd=str(working_dir), env=runtime_env, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        url = f"http://127.0.0.1:{port}"
        deadline = time.monotonic() + timeout
        started = False
        last_error: Exception | None = None
        try:
            while time.monotonic() < deadline:
                if proc.poll() is not None:
                    break
                try:
                    with urllib.request.urlopen(url, timeout=1.0):
                        started = True
                        break
                except Exception as exc:  # pragma: no cover - transient startup
                    last_error = exc
                    time.sleep(0.25)
            if not started and proc.poll() is None:
                last_error = last_error or TimeoutError(f"Streamlit app did not start within {timeout} seconds")
        finally:
            if proc.poll() is None:
                proc.terminate()
                try:
                    stdout, stderr = proc.communicate(timeout=10)
                except subprocess.TimeoutExpired:
                    proc.kill()
                    stdout, stderr = proc.communicate()
            else:
                stdout, stderr = proc.communicate()
        result = CommandResult(
            command=command,
            cwd=working_dir,
            returncode=proc.returncode if proc.returncode is not None else 1,
            stdout=stdout or "",
            stderr=stderr or "",
            duration_seconds=time.perf_counter() - start,
            started_at=started_at,
            finished_at=time.time(),
        )
        if last_error is not None and not started and result.returncode == 0:
            result = CommandResult(
                command=command,
                cwd=working_dir,
                returncode=1,
                stdout=result.stdout,
                stderr=f"{result.stderr}\n{last_error!r}",
                duration_seconds=result.duration_seconds,
                started_at=result.started_at,
                finished_at=result.finished_at,
            )
        return result


__all__ = [
    "CommandResult",
    "ContractScenario",
    "FlowScenario",
    "PipelineScenario",
    "PipelineTester",
    "Scenario",
    "SnapshotScenario",
    "StepScenario",
    "TestMode",
    "TestModeSpec",
    "TestPlan",
    "TestRecipe",
    "Tester",
    "UITester",
]
